#
# LSST Data Management System
# Copyright 2008-2016 AURA/LSST.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <http://www.lsstcorp.org/LegalNotices/>.
#

from __future__ import annotations

__all__ = ["Task", "TaskError"]

import contextlib
import logging
import weakref
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import lsst.utils
import lsst.utils.logging
from lsst.pex.config import ConfigurableField
from lsst.utils.timer import logInfo

if TYPE_CHECKING:
    from lsst.pex.config import Config

try:
    import lsstDebug  # type: ignore
except ImportError:
    lsstDebug = None

from ._task_metadata import TaskMetadata

# This defines the Python type to use for task metadata. It is a private
# class variable that can be accessed by other closely-related middleware
# code and test code.
_TASK_METADATA_TYPE = TaskMetadata
_TASK_FULL_METADATA_TYPE = TaskMetadata


class TaskError(Exception):
    """Use to report errors for which a traceback is not useful.

    Notes
    -----
    Examples of such errors:

    - processCcd is asked to run detection, but not calibration, and no calexp
      is found.
    - coadd finds no valid images in the specified patch.
    """

    pass


class Task:
    r"""Base class for data processing tasks.

    See :ref:`task-framework-overview` to learn what tasks are, and
    :ref:`creating-a-task` for more information about writing tasks.

    Parameters
    ----------
    config : `Task.ConfigClass` instance, optional
        Configuration for this task (an instance of Task.ConfigClass, which
        is a task-specific subclass of `lsst.pex.config.Config`, or `None`.
        If `None`:

        - If parentTask specified then defaults to parentTask.config.\<name>
        - If parentTask is None then defaults to self.ConfigClass()

    name : `str`, optional
        Brief name of task, or `None`; if `None` then defaults to
        `Task._DefaultName`
    parentTask : `Task`-type, optional
        The parent task of this subtask, if any.

        - If `None` (a top-level task) then you must specify config and name
          is ignored.
        - If not `None` (a subtask) then you must specify name.
    log : `logging.Logger` or subclass, optional
        Log whose name is used as a log name prefix, or `None` for no prefix.
        Ignored if is parentTask specified, in which case
        ``parentTask.log``\ 's name is used as a prefix. The task's log name is
        ``prefix + "." + name`` if a prefix exists, else ``name``. The task's
        log is then a child logger of ``parentTask.log`` (if ``parentTask``
        specified), or a child logger of the log from the argument
        (if ``log`` is not `None`).

    Raises
    ------
    RuntimeError
        Raised under these circumstances:

        - If ``parentTask`` is `None` and ``config`` is `None`.
        - If ``parentTask`` is not `None` and ``name`` is `None`.
        - If ``name`` is `None` and ``_DefaultName`` does not exist.

    Notes
    -----
    Useful attributes include:

    - ``log``: an `logging.Logger` or subclass.
    - ``config``: task-specific configuration; an instance of ``ConfigClass``
      (see below).
    - ``metadata``: a `TaskMetadata` for
      collecting task-specific metadata, e.g. data quality and performance
      metrics. This is data that is only meant to be persisted, never to be
      used by the task.

    Subclasses typically have a method named ``runDataRef`` to perform the
    main data processing. Details:

    - ``runDataRef`` should process the minimum reasonable amount of data,
      typically a single CCD.  Iteration, if desired, is performed by a caller
      of the method. This is good design and allows multiprocessing without
      the run method having to support it directly.
    - If ``runDataRef`` can persist or unpersist data:

      - ``runDataRef`` should accept a butler data reference (or a collection
        of data references, if appropriate, e.g. coaddition).
      - There should be a way to run the task without persisting data.
        Typically the run method returns all data, even if it is persisted, and
        the task's config method offers a flag to disable persistence.

    **Deprecated:** Tasks other than cmdLineTask.CmdLineTask%s should *not*
    accept a blob such as a butler data reference.  How we will handle data
    references is still TBD, so don't make changes yet!
    RHL 2014-06-27

    Subclasses must also have an attribute ``ConfigClass`` that is a subclass
    of `lsst.pex.config.Config` which configures the task. Subclasses should
    also have an attribute ``_DefaultName``: the default name if there is no
    parent task. ``_DefaultName`` is required for subclasses of
    `~lsst.pipe.base.CmdLineTask` and recommended for subclasses of Task
    because it simplifies construction (e.g. for unit tests).

    Tasks intended to be run from the command line should be subclasses of
    `~lsst.pipe.base.CmdLineTask` not Task.
    """

    ConfigClass: ClassVar[Type[Config]]
    _DefaultName: ClassVar[str]

    _add_module_logger_prefix: bool = True
    """Control whether the module prefix should be prepended to default
    logger names."""

    def __init__(
        self,
        config: Optional[Config] = None,
        name: Optional[str] = None,
        parentTask: Optional[Task] = None,
        log: Optional[Union[logging.Logger, lsst.utils.logging.LsstLogAdapter]] = None,
    ):
        self.metadata = _TASK_METADATA_TYPE()
        self.__parentTask: Optional[weakref.ReferenceType]
        self.__parentTask = parentTask if parentTask is None else weakref.ref(parentTask)

        if parentTask is not None:
            if name is None:
                raise RuntimeError("name is required for a subtask")
            self._name = name
            self._fullName = parentTask._computeFullName(name)
            if config is None:
                config = getattr(parentTask.config, name)
            self._taskDict: Dict[str, weakref.ReferenceType[Task]] = parentTask._taskDict
            loggerName = parentTask.log.getChild(name).name
        else:
            if name is None:
                name = getattr(self, "_DefaultName", None)
                if name is None:
                    raise RuntimeError("name is required for a task unless it has attribute _DefaultName")
                name = self._DefaultName
            self._name = name
            self._fullName = self._name
            if config is None:
                config = self.ConfigClass()
            self._taskDict = dict()
            loggerName = self._fullName
            if log is not None and log.name:
                loggerName = log.getChild(loggerName).name
            elif self._add_module_logger_prefix:
                # Prefix the logger name with the root module name.
                # We want all Task loggers to have this prefix to make
                # it easier to control them. This can be disabled by
                # a Task setting the class property _add_module_logger_prefix
                # to False -- in which case the logger name will not be
                # modified.
                module_name = self.__module__
                module_root = module_name.split(".")[0] + "."
                if not loggerName.startswith(module_root):
                    loggerName = module_root + loggerName

        # Get a logger (that might be a subclass of logging.Logger).
        self.log: lsst.utils.logging.LsstLogAdapter = lsst.utils.logging.getLogger(loggerName)
        self.config: Config = config
        if lsstDebug:
            self._display = lsstDebug.Info(self.__module__).display
        else:
            self._display = None
        self._taskDict[self._fullName] = weakref.ref(self)

    @property
    def _parentTask(self) -> Optional[Task]:
        return self.__parentTask if self.__parentTask is None else self.__parentTask()

    def emptyMetadata(self) -> None:
        """Empty (clear) the metadata for this Task and all sub-Tasks."""
        for wref in self._taskDict.values():
            subtask = wref()
            assert subtask is not None, "Unexpected garbage collection of subtask."
            subtask.metadata = _TASK_METADATA_TYPE()

    # We use Any instead of lsst.afw.table.BaseCatalog here to avoid a
    # type-only dependency on afw.  It's unclear whether this will survive
    # Gen2 anyway, or how we might adapt it to work with non-afw catalogs
    # (e.g. Parquet).
    def getSchemaCatalogs(self) -> Dict[str, Any]:
        """Get the schemas generated by this task.

        Returns
        -------
        schemaCatalogs : `dict`
            Keys are butler dataset type, values are an empty catalog (an
            instance of the appropriate `lsst.afw.table` Catalog type) for
            this task.

        Notes
        -----

        .. warning::

           Subclasses that use schemas must override this method. The default
           implementation returns an empty dict.

        This method may be called at any time after the Task is constructed,
        which means that all task schemas should be computed at construction
        time, *not* when data is actually processed. This reflects the
        philosophy that the schema should not depend on the data.

        Returning catalogs rather than just schemas allows us to save e.g.
        slots for SourceCatalog as well.

        See also
        --------
        Task.getAllSchemaCatalogs
        """
        return {}

    def getAllSchemaCatalogs(self) -> Dict[str, Any]:
        """Get schema catalogs for all tasks in the hierarchy, combining the
        results into a single dict.

        Returns
        -------
        schemacatalogs : `dict`
            Keys are butler dataset type, values are a empty catalog (an
            instance of the appropriate `lsst.afw.table` Catalog type) for all
            tasks in the hierarchy, from the top-level task down
            through all subtasks.

        Notes
        -----
        This method may be called on any task in the hierarchy; it will return
        the same answer, regardless.

        The default implementation should always suffice. If your subtask uses
        schemas the override `Task.getSchemaCatalogs`, not this method.
        """
        schemaDict = self.getSchemaCatalogs()
        for wref in self._taskDict.values():
            subtask = wref()
            assert subtask is not None, "Unexpected garbage collection of subtask."
            schemaDict.update(subtask.getSchemaCatalogs())
        return schemaDict

    def getFullMetadata(self) -> TaskMetadata:
        """Get metadata for all tasks.

        Returns
        -------
        metadata : `TaskMetadata`
            The keys are the full task name.
            Values are metadata for the top-level task and all subtasks,
            sub-subtasks, etc.

        Notes
        -----
        The returned metadata includes timing information (if
        ``@timer.timeMethod`` is used) and any metadata set by the task. The
        name of each item consists of the full task name with ``.`` replaced
        by ``:``, followed by ``.`` and the name of the item, e.g.::

            topLevelTaskName:subtaskName:subsubtaskName.itemName

        using ``:`` in the full task name disambiguates the rare situation
        that a task has a subtask and a metadata item with the same name.
        """
        fullMetadata = _TASK_FULL_METADATA_TYPE()
        for fullName, wref in self.getTaskDict().items():
            subtask = wref()
            assert subtask is not None, "Unexpected garbage collection of subtask."
            fullMetadata[fullName.replace(".", ":")] = subtask.metadata
        return fullMetadata

    def getFullName(self) -> str:
        """Get the task name as a hierarchical name including parent task
        names.

        Returns
        -------
        fullName : `str`
            The full name consists of the name of the parent task and each
            subtask separated by periods. For example:

            - The full name of top-level task "top" is simply "top".
            - The full name of subtask "sub" of top-level task "top" is
              "top.sub".
            - The full name of subtask "sub2" of subtask "sub" of top-level
              task "top" is "top.sub.sub2".
        """
        return self._fullName

    def getName(self) -> str:
        """Get the name of the task.

        Returns
        -------
        taskName : `str`
            Name of the task.

        See also
        --------
        getFullName
        """
        return self._name

    def getTaskDict(self) -> Dict[str, weakref.ReferenceType[Task]]:
        """Get a dictionary of all tasks as a shallow copy.

        Returns
        -------
        taskDict : `dict`
            Dictionary containing full task name: task object for the top-level
            task and all subtasks, sub-subtasks, etc.
        """
        return self._taskDict.copy()

    def makeSubtask(self, name: str, **keyArgs: Any) -> None:
        """Create a subtask as a new instance as the ``name`` attribute of this
        task.

        Parameters
        ----------
        name : `str`
            Brief name of the subtask.
        keyArgs
            Extra keyword arguments used to construct the task. The following
            arguments are automatically provided and cannot be overridden:

            - "config".
            - "parentTask".

        Notes
        -----
        The subtask must be defined by ``Task.config.name``, an instance of
        `~lsst.pex.config.ConfigurableField` or
        `~lsst.pex.config.RegistryField`.
        """
        taskField = getattr(self.config, name, None)
        if taskField is None:
            raise KeyError(f"{self.getFullName()}'s config does not have field {name!r}")
        subtask = taskField.apply(name=name, parentTask=self, **keyArgs)
        setattr(self, name, subtask)

    @contextlib.contextmanager
    def timer(self, name: str, logLevel: int = logging.DEBUG) -> Iterator[None]:
        """Context manager to log performance data for an arbitrary block of
        code.

        Parameters
        ----------
        name : `str`
            Name of code being timed; data will be logged using item name:
            ``Start`` and ``End``.
        logLevel
            A `logging` level constant.

        Examples
        --------
        Creating a timer context:

        .. code-block:: python

            with self.timer("someCodeToTime"):
                pass  # code to time

        See also
        --------
        timer.logInfo
        """
        logInfo(obj=self, prefix=name + "Start", logLevel=logLevel)
        try:
            yield
        finally:
            logInfo(obj=self, prefix=name + "End", logLevel=logLevel)

    @classmethod
    def makeField(cls, doc: str) -> ConfigurableField:
        """Make a `lsst.pex.config.ConfigurableField` for this task.

        Parameters
        ----------
        doc : `str`
            Help text for the field.

        Returns
        -------
        configurableField : `lsst.pex.config.ConfigurableField`
            A `~ConfigurableField` for this task.

        Examples
        --------
        Provides a convenient way to specify this task is a subtask of another
        task.

        Here is an example of use:

        .. code-block:: python

            class OtherTaskConfig(lsst.pex.config.Config):
                aSubtask = ATaskClass.makeField("brief description of task")
        """
        return ConfigurableField(doc=doc, target=cls)

    def _computeFullName(self, name: str) -> str:
        """Compute the full name of a subtask or metadata item, given its brief
        name.

        Parameters
        ----------
        name : `str`
            Brief name of subtask or metadata item.

        Returns
        -------
        fullName : `str`
            The full name: the ``name`` argument prefixed by the full task name
            and a period.

        Notes
        -----
        For example: if the full name of this task is "top.sub.sub2"
        then ``_computeFullName("subname")`` returns
        ``"top.sub.sub2.subname"``.
        """
        return f"{self._fullName}.{name}"

    @staticmethod
    def _unpickle_via_factory(
        factory: Callable[..., Task], args: Sequence[Any], kwargs: Dict[str, Any]
    ) -> Task:
        """Unpickle something by calling a factory

        Allows subclasses to unpickle using `__reduce__` with keyword
        arguments as well as positional arguments.
        """
        return factory(*args, **kwargs)

    def _reduce_kwargs(self) -> Dict[str, Any]:
        """Returns a dict of the keyword arguments that should be used
        by `__reduce__`.

        Subclasses with additional arguments should always call the parent
        class method to ensure that the standard parameters are included.

        Returns
        -------
        kwargs : `dict`
            Keyword arguments to be used when pickling.
        """
        return dict(
            config=self.config,
            name=self._name,
            parentTask=self._parentTask,
        )

    def __reduce__(
        self,
    ) -> Tuple[
        Callable[[Callable[..., Task], Sequence[Any], Dict[str, Any]], Task],
        Tuple[Type[Task], Sequence[Any], Dict[str, Any]],
    ]:
        """Pickler."""
        return self._unpickle_via_factory, (self.__class__, [], self._reduce_kwargs())



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

"""Module Defining variants for valid values used to constrain datasets in a
graph building query.
"""

__all__ = ("DatasetQueryConstraintVariant",)

import warnings
from typing import Iterable, Iterator, Protocol, Type


class DatasetQueryConstraintVariant(Iterable, Protocol):
    """This class is the base for all the valid variants for controling
    constraining graph building queries based on dataset type existence.

    ALL variant corresponds to using all input dataset types to constrain
    a query.

    OFF variant corrresponds to not using any dataset types to constrain a
    graph building query.

    LIST variant should be used when one or more specific names should be used
    in constraining a graph building query.

    Normally the ALL and OFF variants are used as as Singletons, attempting to
    instantiate them (i.e. ALL()) will return in singleton class itself.

    LIST is used as a constructor to the contents (i.e. List(['a', 'b'])).
    Using the LIST variant directly as a singleton will behave the same as if
    it were an empty instance.

    Variants can be directly used, or automatically be selected by using the
    `fromExpression` class method given a valid string.
    """

    ALL: "Type[_ALL]"
    OFF: "Type[_OFF]"
    LIST: "Type[_LIST]"

    @classmethod
    def __subclasshook__(cls, subclass):
        if subclass == cls.ALL or subclass == cls.OFF or subclass == cls.LIST:
            return True
        return False

    @classmethod
    def fromExpression(cls, expression: str) -> "DatasetQueryConstraintVariant":
        """Select and return the correct Variant that corresponds to the input
        expression.

        Valid values are `all` for all inputs dataset types in pipeline, `off`
        to not consider dataset type existence as a constraint, single or comma
        seperated list of dataset type names.
        """
        if not isinstance(expression, str):
            raise ValueError("Expression must be a string")
        elif expression == "all":
            return cls.ALL
        elif expression == "off":
            return cls.OFF
        else:
            if " " in expression:
                warnings.warn("Witespace found in expression will be trimmed", RuntimeWarning)
                expression = expression.replace(" ", "")
            members = expression.split(",")
            return cls.LIST(members)


class _ALLMETA(DatasetQueryConstraintVariant, type(Protocol)):
    def __iter__(self) -> Iterator:
        raise NotImplementedError("This variant cannot be iteratted")


class _ALL(metaclass=_ALLMETA):
    def __new__(cls):
        return cls


class _OFFMETA(DatasetQueryConstraintVariant, type(Protocol)):
    def __iter__(self) -> Iterator:
        raise NotImplementedError("This variant cannot be iteratted")


class _OFF(metaclass=_OFFMETA):
    def __new__(cls):
        return cls


class _LISTMETA(type(Protocol)):
    def __iter__(self):
        return iter(tuple())

    def __len__(self):
        return 0

    def __eq__(self, o: object) -> bool:
        if isinstance(o, self):
            return True
        return super().__eq__(o)


class _LIST(DatasetQueryConstraintVariant, metaclass=_LISTMETA):
    def __init__(self, members: Iterable[str]):
        self.members = list(members)

    def __len__(self) -> int:
        return len(self.members)

    def __iter__(self) -> Iterable[str]:
        return iter(self.members)

    def __repr__(self) -> str:
        return repr(self.members)

    def __str__(self) -> str:
        return str(self.members)

    def __eq__(self, o: object) -> bool:
        if isinstance(o, type(self)):
            return self.members == o.members
        return super().__eq__(o)


def suppressInit(self):
    raise NotImplementedError(
        "DatasetQueryConstraintVariants cannot be directly instantiated. "
        "Please use the variants or the fromExpression class method"
    )


DatasetQueryConstraintVariant.__init__ = suppressInit
DatasetQueryConstraintVariant.ALL = _ALL
DatasetQueryConstraintVariant.OFF = _OFF
DatasetQueryConstraintVariant.LIST = _LIST



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Shim classes that provide (limited) Gen2 Butler interfaces to Gen3
data repositories.

All of the classes here still operate on Gen3 data IDs - the shim layer
essentially assumes that Gen2 Tasks treat the data ID as an opaque blob, which
is usually (but not always true).  When it isn't, the best approach now is
probably to use the GENERATION class attribute on the Butler classes to
special-case code for each generation.
"""

__all__ = ("ShimButler", "ShimButlerSubset", "ShimDataRef")

from lsst.daf.butler import StorageClassFactory
from lsst.daf.persistence import NoResults


class ShimButler:
    """A shim for a Gen2 `~lsst.daf.persistence.Butler` with a Gen3
    `~lsst.daf.butler.Butler` backend.

    Parameters
    ----------
    butler3 : `lsst.daf.butler.Butler`
        Generation 3 Butler instance.
    """

    GENERATION = 2.5
    """This is a Generation 2 shim for a Generation3 Butler.
    """

    def __init__(self, butler3):
        self._butler3 = butler3

    def _makeDataId(self, dataId=None, **rest):
        """Construct a full data ID by merging the given arguments with the
        internal data ID.
        """
        fullDataId = dict()
        if dataId is not None:
            fullDataId.update(dataId)
        fullDataId.update(rest)
        return fullDataId

    def _translateDatasetType(self, datasetType):
        if "_" in datasetType:
            if datasetType.endswith("_md"):
                return f"{datasetType[:-3]}.metadata"
            for component in StorageClassFactory().getStorageClass("Exposure").components:
                suffix = f"_{component}"
                if datasetType.endswith(suffix):
                    return "{}.{}".format(datasetType[: -len(suffix)], component)
        return datasetType

    def datasetExists(self, datasetType, dataId=None, write=False, **rest):
        """Check whether a datataset exists in the repository.

        Parameters
        ----------
        datasetType : `str`
            Name of the Gen2 dataset type.
        dataId : `dict` or `~lsst.daf.butler.DataId`, optional
            A Generation 3 data ID that identifies the dataset.
        write : `bool`
            This option is provided for compatibility with
            `lsst.daf.persistence.Butler`, but must be `False`.
        rest
            Additional key-value pairs to augment the given data ID.

        Returns
        -------
        exists : `bool`
            `True` if the dataset is present in the repository, `False`
            otherwise.
        """
        if write:
            raise NotImplementedError("ShimButler cannot implement datasetExists with 'write=True'")
        datasetType = self._translateDatasetType(datasetType)
        try:
            return self._butler3.datasetExists(datasetType, self._makeDataId(dataId, **rest))
        except LookupError:
            # Gen3 datasetExists raises if Dataset is not present in Registry;
            # Gen2 does not distinguish between present in Datastore and
            # present in Registry.
            return False

    def get(self, datasetType, dataId=None, immediate=True, **rest):
        """Retrieve a dataset.

        Parameters
        ----------
        datasetType : `str`
            Name of the Gen2 dataset type.
        dataId : `dict` or `~lsst.daf.butler.DataId`, optional
            A Generation 3 data ID that identifies the dataset.
        immediate : `bool`
            This option is provided for compatibility with
            `lsst.daf.persistence.Butler`, but is ignored.
        rest
            Additional key-value pairs to augment the given data ID.

        Returns
        -------
        dataset
            Retrieved object.

        Raises
        ------
        `~lsst.daf.persistence.NoResults`
            Raised if the dataset does not exist.
        """
        datasetType = self._translateDatasetType(datasetType)
        fullDataId = self._makeDataId(dataId, **rest)
        if datasetType.endswith("_sub"):
            import lsst.afw.image

            datasetType = datasetType[: -len("_sub")]
            parameters = dict(bbox=fullDataId.pop("bbox"))
            origin = fullDataId.pop("imageOrigin", lsst.afw.image.PARENT)
            parameters["origin"] = origin
        else:
            parameters = {}
        try:
            return self._butler3.get(datasetType, fullDataId, parameters=parameters)
        except (FileNotFoundError, LookupError) as err:
            raise NoResults(str(err), datasetType, fullDataId)

    def put(self, obj, datasetType, dataId=None, doBackup=False, **rest):
        """Write a dataset.

        Parameters
        ----------
        obj
            Object to write.
        datasetType : `str`
            Name of the Gen2 dataset type.
        dataId : `dict` or `~lsst.daf.butler.DataId`, optional
            A Generation 3 data ID that identifies the dataset.
        doBackup : `bool`
            This option is provided for compatibility with
            `lsst.daf.persistence.Butler`, but must be `False`.
        rest
            Additional key-value pairs to augment the given data ID.
        """
        if doBackup:
            raise NotImplementedError("ShimButler cannot implement put with 'doBackup=True'")
        datasetType = self._translateDatasetType(datasetType)
        self._butler3.put(obj, datasetType, self._makeDataId(dataId, **rest))

    def dataRef(self, datasetType, level=None, dataId=None, **rest):
        """Return a DataRef associated with the given dataset type and data ID.

        Parameters
        ----------
        datasetType : `str`
            Name of the dataset type.
        dataId : `dict` or `~lsst.daf.butler.DataId`, optional
            A Generation 3 data ID that identifies the dataset.
        level
            This option is provided for compatibility with
            `lsst.daf.persistence.Butler`, but must be `None`.
        rest
            Additional key-value pairs to augment the given data ID.
        """
        if level is not None:
            raise NotImplementedError("ShimButler cannot implement dataRef with 'level != None'")
        fullDataId = {}
        if dataId is not None:
            fullDataId.update(dataId)
        fullDataId.update(rest)
        return next(iter(ShimButlerSubset(self, datasetType, [fullDataId])))


class ShimButlerSubset:
    """A shim for a Gen2 `~lsst.daf.persistence.ButlerSubset` with a Gen3
    `~lsst.daf.butler.Butler` backend.

    Parameters
    ----------
    butler : `ShimButler`
        Butler shim instance.
    datasetType : `str`
        Name of the dataset type.
    dataIds : iterable of `dict` or `~lsst.daf.butler.DataId`
        Generation 3 data IDs that define the data in this subset.
    """

    GENERATION = 2.5
    """This is a Generation 2 shim for a Generation3 Butler.
    """

    def __init__(self, butler, datasetType, dataIds):
        self.butler = butler
        self.datasetType = datasetType
        self._dataIds = tuple(dataIds)

    def __len__(self):
        return len(self._dataIds)

    def __iter__(self):
        for dataId in self._dataIds:
            yield ShimDataRef(self, dataId)


class ShimDataRef:
    """A shim for a Gen2 `~lsst.daf.persistence.ButlerDataRef` with a Gen3
    `~lsst.daf.butler.Butler` backend.

    Parameters
    ----------
    butlerSubset : `ShimButlerSubset`
        ButlerSubset shim instance.  Sets the butler and default dataset type
        used by the Dataref.
    dataId : `dict` or `~lsst.daf.butler.DataId`
        Generation 3 data ID associated with this reference.
    """

    GENERATION = 2.5
    """This is a Generation 2 shim for a Generation3 Butler.
    """

    def __init__(self, butlerSubset, dataId):
        self.butlerSubset = butlerSubset
        self.dataId = dataId

    def get(self, datasetType=None, **rest):
        """Retrieve a dataset.

        Parameters
        ----------
        datasetType : `str`, optional.
            Name of the dataset type.  Defaults to the dataset type used to
            construct the `ShimButlerSubset`.
        rest
            Additional arguments forwarded to `ShimButler.get`.

        Returns
        -------
        dataset
            Retrieved object.

        Raises
        ------
        `~lsst.daf.persistence.NoResults`
            Raised if the dataset does not exist.
        """
        if datasetType is None:
            datasetType = self.butlerSubset.datasetType
        return self.butlerSubset.butler.get(datasetType, self.dataId, **rest)

    def put(self, obj, datasetType=None, doBackup=False, **rest):
        """Write a dataset.

        Parameters
        ----------
        obj
            Object to write.
        datasetType : `str`, optional
            Name of the dataset type.  Defaults to the dataset type used to
            construct the `ShimButlerSubset`.
        doBackup : `bool`
            This option is provided for compatibility with
            `lsst.daf.persistence.ButlerDataRef`, but must be `False`.
        rest
            Additional arguments forwarded to `ShimButler.put`.
        """
        if datasetType is None:
            datasetType = self.butlerSubset.datasetType
        self.butlerSubset.butler.put(obj, datasetType, self.dataId, doBackup=doBackup, **rest)

    def datasetExists(self, datasetType=None, write=False, **rest):
        """Check whether a datataset exists in the repository.

        Parameters
        ----------
        datasetType : `str`, optional
            Name of the dataset type.  Defaults to the dataset type used to
            construct the `ShimButlerSubset`.
        write : `bool`
            This option is provided for compatibility with
            `lsst.daf.persistence.ButlerDataRef`, but must be `False`.
        rest
            Additional arguments forwarded to `ShimButler.datasetExists`.

        Returns
        -------
        exists : `bool`
            `True` if the dataset is present in the repository, `False`
            otherwise.
        """
        if datasetType is None:
            datasetType = self.butlerSubset.datasetType
        return self.butlerSubset.butler.datasetExists(datasetType, self.dataId, write=write, **rest)

    def getButler(self):
        """Return the (shim) Butler used by this DataRef."""
        return self.butlerSubset.butler



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ("TRACE", "VERBOSE", "getTaskLogger")

from deprecated.sphinx import deprecated
from lsst.utils.logging import TRACE, VERBOSE, getLogger


@deprecated(
    reason="getTaskLogger has been replaced by lsst.utils.logging.getLogger. Will be removed after v24.",
    version="v24",
    category=FutureWarning,
)
def getTaskLogger(name=None, logger=None):
    """Get a logger compatible with LSST usage.

    Parameters
    ----------
    name : `str`, optional
        Name of the logger. Root logger if `None`.
    logger : `logging.Logger`
        If given the logger is converted to the relevant logger class.
        If ``name`` is given the logger is assumed to be a child of the
        supplied logger.

    Returns
    -------
    logger : `lsst.utils.logging.LsstLogAdapter`
        The relevant logger.

    Notes
    -----
    A `logging.LoggerAdapter` is used since it is easier to provide a more
    uniform interface than when using `logging.setLoggerClass`. An adapter
    can be wrapped around the root logger and the `~logging.setLoggerClass`
    will return the logger first given that name even if the name was
    used before the `Task` was created.
    """
    return getLogger(name=name, logger=logger)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for ResourceConfig.
"""

import unittest

import lsst.pex.config as pexConfig
import lsst.pipe.base as pipeBase
import lsst.utils.tests
from lsst.daf.butler import StorageClass, StorageClassFactory


class NullConnections(pipeBase.PipelineTaskConnections, dimensions=()):
    pass


class NoResourceTask(pipeBase.PipelineTask):
    _DefaultName = "no_resource_task"
    ConfigClass = pipeBase.PipelineTaskConfig


class OneConfig(pipeBase.PipelineTaskConfig, pipelineConnections=NullConnections):
    resources = pexConfig.ConfigField(dtype=pipeBase.ResourceConfig, doc="Resource configuration")


class OneTask(pipeBase.PipelineTask):
    _DefaultName = "one_task"
    ConfigClass = OneConfig


class TwoConfig(pipeBase.PipelineTaskConfig, pipelineConnections=NullConnections):
    resources = pexConfig.ConfigField(dtype=pipeBase.ResourceConfig, doc="Resource configuration")

    def setDefaults(self):
        self.resources.minMemoryMB = 1024
        self.resources.minNumCores = 32


class TwoTask(pipeBase.PipelineTask):
    _DefaultName = "two_task"
    ConfigClass = TwoConfig


class TaskTestCase(unittest.TestCase):
    """A test case for Task"""

    @classmethod
    def setUpClass(cls):
        for name in ("SCA", "SCB", "SCC", "SCX", "SCY"):
            StorageClassFactory().registerStorageClass(StorageClass(name))

    def testNoResource(self):
        """Test for a task without resource config"""
        task = NoResourceTask()
        res_config = task.getResourceConfig()
        self.assertIs(res_config, None)

    def testOneResource(self):
        """Test for a task with resource config"""
        task = OneTask()
        res_config = task.getResourceConfig()
        self.assertIsNot(res_config, None)
        self.assertIs(res_config.minMemoryMB, None)
        self.assertEqual(res_config.minNumCores, 1)

    def testTwoResource(self):
        """Test for a task with resource config and special defaults"""
        task = TwoTask()
        res_config = task.getResourceConfig()
        self.assertIsNot(res_config, None)
        self.assertEqual(res_config.minMemoryMB, 1024)
        self.assertEqual(res_config.minNumCores, 32)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    pass


def setup_module(module):
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



#
# LSST Data Management System
# Copyright 2008, 2009, 2010 LSST Corporation.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <http://www.lsstcorp.org/LegalNotices/>.
#
import unittest

import lsst.pipe.base as pipeBase
import lsst.utils.tests


class StructTestCase(unittest.TestCase):
    """A test case for Struct"""

    def setUp(self):
        self.valDict = dict(
            foo=1,
            bar=(1, 2, 3),
            baz="value for baz",
            alist=[3, 5, 7, 9],
        )

    def tearDown(self):
        self.valDict = None

    def testInit(self):
        """Test Struct.__init__"""
        s = pipeBase.Struct(**self.valDict)
        self.assertEqual(self.valDict, s.getDict())

        for name, val in self.valDict.items():
            self.assertEqual(getattr(s, name), val)

    def testInitException(self):
        """Test that struct key names cannot start with double underscores."""
        with self.assertRaises(RuntimeError):
            pipeBase.Struct(__foo=13)

    def testSet(self):
        """Test adding values via struct.name=val"""
        s = pipeBase.Struct()
        for name, val in self.valDict.items():
            setattr(s, name, val)

        self.assertEqual(self.valDict, s.getDict())

    def testCopy(self):
        """Test copy, which returns a shallow copy"""
        s = pipeBase.Struct(**self.valDict)
        sc = s.copy()
        self.assertEqual(s.getDict(), sc.getDict())

        # shallow copy, so changing a list should propagate (not necessarily a
        # feature)
        sc.alist[0] = 97
        self.assertEqual(s, sc)

        sc.foo += 1
        self.assertNotEqual(s, sc)

    def testMergeItems(self):
        """Test mergeItems"""
        s = pipeBase.Struct(**self.valDict)
        newS = pipeBase.Struct()
        newS.mergeItems(s)
        # with no names listed, should merge nothing
        self.assertEqual(len(newS), 0)
        self.assertNotEqual(s, newS)

        newS.mergeItems(s, "foo", "bar")
        self.assertEqual(len(newS), 2)
        self.assertNotEqual(s, newS)

        newS.mergeItems(s, "baz", "alist")
        self.assertEqual(len(newS), 4)

        for name, val in newS.getDict().items():
            self.assertEqual(val, self.valDict[name])
            with self.assertRaises(RuntimeError):
                newS.mergeItems(s, name)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    pass


def setup_module(module):
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



#
# LSST Data Management System
# Copyright 2013 LSST Corporation.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <http://www.lsstcorp.org/LegalNotices/>.
#
import io
import sys
import textwrap
import unittest

import lsst.pex.config as pexConfig
import lsst.pipe.base as pipeBase

try:
    import lsst.pipe.base.argumentParser as argumentParser
except ImportError:
    argumentParser = None
import lsst.utils.tests


class SimpleTaskConfig(pexConfig.Config):
    ff3 = pexConfig.Field(doc="float 1", dtype=float, default=3.1)
    sf3 = pexConfig.Field(doc="str 1", dtype=str, default="default for sf1")


class SimpleTask(pipeBase.Task):
    ConfigClass = SimpleTaskConfig


class TaskWithSubtasksConfig(pexConfig.Config):
    sst1 = SimpleTask.makeField(doc="sub-subtask 1")
    sst2 = SimpleTask.makeField(doc="sub-subtask 2")
    ff2 = pexConfig.Field(doc="float 1", dtype=float, default=3.1)
    sf2 = pexConfig.Field(doc="str 1", dtype=str, default="default for sf1")


class TaskWithSubtasks(pipeBase.Task):
    ConfigClass = TaskWithSubtasksConfig


class MainTaskConfig(pexConfig.Config):
    st1 = TaskWithSubtasks.makeField(doc="subtask 1")
    st2 = TaskWithSubtasks.makeField(doc="subtask 2")
    ff1 = pexConfig.Field(doc="float 2", dtype=float, default=3.1)
    sf1 = pexConfig.Field(doc="str 2", dtype=str, default="default for strField")


class MainTask(pipeBase.Task):
    ConfigClass = MainTaskConfig


c = MainTaskConfig()


@unittest.skipIf(argumentParser is None, "Gen2 argument parser is not available")
class ShowTasksTestCase(unittest.TestCase):
    """A test case for the code that implements ArgumentParser's --show tasks
    option.
    """

    def testBasicShowTaskHierarchy(self):
        """Test basic usage of show"""
        config = MainTaskConfig()
        expectedData = """
        Subtasks:
        st1: {0}.TaskWithSubtasks
        st1.sst1: {0}.SimpleTask
        st1.sst2: {0}.SimpleTask
        st2: {0}.TaskWithSubtasks
        st2.sst1: {0}.SimpleTask
        st2.sst2: {0}.SimpleTask
        """.format(
            __name__
        )
        tempStdOut = io.StringIO()
        savedStdOut, sys.stdout = sys.stdout, tempStdOut
        try:
            argumentParser.showTaskHierarchy(config)
        finally:
            sys.stdout = savedStdOut
        formatRead = tempStdOut.getvalue().strip()
        formatExpected = textwrap.dedent(expectedData).strip()
        self.assertEqual(formatRead, formatExpected)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    pass


def setup_module(module):
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

"""Module defining GraphBuilder class and related methods.
"""

__all__ = ["GraphBuilder"]

# -------------------------------
#  Imports of standard modules --
# -------------------------------
import itertools
import logging
from collections import ChainMap
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Collection, Dict, Iterable, Iterator, List, Mapping, Optional, Set, Union

from lsst.daf.butler import (
    CollectionSearch,
    CollectionType,
    DataCoordinate,
    DatasetRef,
    DatasetType,
    Datastore,
    DatastoreRecordData,
    DimensionGraph,
    DimensionUniverse,
    NamedKeyDict,
    NamedValueSet,
    Quantum,
    Registry,
)
from lsst.daf.butler.registry.queries import DataCoordinateQueryResults
from lsst.utils import doImportType

from ._datasetQueryConstraints import DatasetQueryConstraintVariant
from ._status import NoWorkFound

# -----------------------------
#  Imports for other modules --
# -----------------------------
from .connections import AdjustQuantumHelper, iterConnections
from .graph import QuantumGraph
from .pipeline import Pipeline, PipelineDatasetTypes, TaskDatasetTypes, TaskDef

# ----------------------------------
#  Local non-exported definitions --
# ----------------------------------

_LOG = logging.getLogger(__name__)


class _DatasetDict(NamedKeyDict[DatasetType, Dict[DataCoordinate, DatasetRef]]):
    """A custom dictionary that maps `DatasetType` to a nested dictionary of
    the known `DatasetRef` instances of that type.

    Parameters
    ----------
    args
        Positional arguments are forwarded to the `dict` constructor.
    universe : `DimensionUniverse`
        Universe of all possible dimensions.
    """

    def __init__(self, *args: Any, universe: DimensionUniverse):
        super().__init__(*args)
        self.universe = universe

    @classmethod
    def fromDatasetTypes(
        cls, datasetTypes: Iterable[DatasetType], *, universe: DimensionUniverse
    ) -> _DatasetDict:
        """Construct a dictionary from a flat iterable of `DatasetType` keys.

        Parameters
        ----------
        datasetTypes : `iterable` of `DatasetType`
            DatasetTypes to use as keys for the dict.  Values will be empty
            dictionaries.
        universe : `DimensionUniverse`
            Universe of all possible dimensions.

        Returns
        -------
        dictionary : `_DatasetDict`
            A new `_DatasetDict` instance.
        """
        return cls({datasetType: {} for datasetType in datasetTypes}, universe=universe)

    @classmethod
    def fromSubset(
        cls, datasetTypes: Collection[DatasetType], first: _DatasetDict, *rest: _DatasetDict
    ) -> _DatasetDict:
        """Return a new dictionary by extracting items corresponding to the
        given keys from one or more existing dictionaries.

        Parameters
        ----------
        datasetTypes : `iterable` of `DatasetType`
            DatasetTypes to use as keys for the dict.  Values will be obtained
            by lookups against ``first`` and ``rest``.
        first : `_DatasetDict`
            Another dictionary from which to extract values.
        rest
            Additional dictionaries from which to extract values.

        Returns
        -------
        dictionary : `_DatasetDict`
            A new dictionary instance.
        """
        combined = ChainMap(first, *rest)

        # Dataset types known to match immediately can be processed
        # without checks.
        matches = combined.keys() & set(datasetTypes)
        _dict = {k: combined[k] for k in matches}

        if len(_dict) < len(datasetTypes):
            # Work out which ones are missing.
            missing_datasetTypes = set(datasetTypes) - _dict.keys()

            # Get the known names for comparison.
            combined_by_name = {k.name: k for k in combined}

            missing = set()
            incompatible = {}
            for datasetType in missing_datasetTypes:
                # The dataset type is not found. It may not be listed
                # or it may be that it is there with the same name
                # but different definition.
                if datasetType.name in combined_by_name:
                    # This implies some inconsistency in definitions
                    # for connections. If there is support for storage
                    # class conversion we can let it slide.
                    # At this point we do not know
                    # where the inconsistency is but trust that down
                    # stream code will be more explicit about input
                    # vs output incompatibilities.
                    existing = combined_by_name[datasetType.name]
                    if existing.is_compatible_with(datasetType) or datasetType.is_compatible_with(existing):
                        _LOG.warning(
                            "Dataset type mismatch (%s != %s) but continuing since they are compatible",
                            datasetType,
                            existing,
                        )
                        _dict[datasetType] = combined[existing]
                    else:
                        incompatible[datasetType] = existing
                else:
                    missing.add(datasetType)

            if missing or incompatible:
                reasons = []
                if missing:
                    reasons.append(
                        "DatasetTypes {'.'.join(missing)} not present in list of known types: "
                        + ", ".join(d.name for d in combined)
                    )
                if incompatible:
                    for x, y in incompatible.items():
                        reasons.append(f"{x} incompatible with {y}")
                raise KeyError("Errors matching dataset types: " + " & ".join(reasons))

        return cls(_dict, universe=first.universe)

    @property
    def dimensions(self) -> DimensionGraph:
        """The union of all dimensions used by all dataset types in this
        dictionary, including implied dependencies (`DimensionGraph`).
        """
        base = self.universe.empty
        if len(self) == 0:
            return base
        return base.union(*[datasetType.dimensions for datasetType in self.keys()])

    def unpackSingleRefs(self) -> NamedKeyDict[DatasetType, DatasetRef]:
        """Unpack nested single-element `DatasetRef` dicts into a new
        mapping with `DatasetType` keys and `DatasetRef` values.

        This method assumes that each nest contains exactly one item, as is the
        case for all "init" datasets.

        Returns
        -------
        dictionary : `NamedKeyDict`
            Dictionary mapping `DatasetType` to `DatasetRef`, with both
            `DatasetType` instances and string names usable as keys.
        """

        def getOne(refs: Dict[DataCoordinate, DatasetRef]) -> DatasetRef:
            (ref,) = refs.values()
            return ref

        return NamedKeyDict({datasetType: getOne(refs) for datasetType, refs in self.items()})

    def unpackMultiRefs(self) -> NamedKeyDict[DatasetType, List[DatasetRef]]:
        """Unpack nested multi-element `DatasetRef` dicts into a new
        mapping with `DatasetType` keys and `set` of `DatasetRef` values.

        Returns
        -------
        dictionary : `NamedKeyDict`
            Dictionary mapping `DatasetType` to `list` of `DatasetRef`, with
            both `DatasetType` instances and string names usable as keys.
        """
        return NamedKeyDict({datasetType: list(refs.values()) for datasetType, refs in self.items()})

    def extract(self, datasetType: DatasetType, dataIds: Iterable[DataCoordinate]) -> Iterator[DatasetRef]:
        """Iterate over the contained `DatasetRef` instances that match the
        given `DatasetType` and data IDs.

        Parameters
        ----------
        datasetType : `DatasetType`
            Dataset type to match.
        dataIds : `Iterable` [ `DataCoordinate` ]
            Data IDs to match.

        Returns
        -------
        refs : `Iterator` [ `DatasetRef` ]
            DatasetRef instances for which ``ref.datasetType == datasetType``
            and ``ref.dataId`` is in ``dataIds``.
        """
        refs = self[datasetType]
        return (refs[dataId] for dataId in dataIds)


class _QuantumScaffolding:
    """Helper class aggregating information about a `Quantum`, used when
    constructing a `QuantumGraph`.

    See `_PipelineScaffolding` for a top-down description of the full
    scaffolding data structure.

    Parameters
    ----------
    task : _TaskScaffolding
        Back-reference to the helper object for the `PipelineTask` this quantum
        represents an execution of.
    dataId : `DataCoordinate`
        Data ID for this quantum.
    """

    def __init__(self, task: _TaskScaffolding, dataId: DataCoordinate):
        self.task = task
        self.dataId = dataId
        self.inputs = _DatasetDict.fromDatasetTypes(task.inputs.keys(), universe=dataId.universe)
        self.outputs = _DatasetDict.fromDatasetTypes(task.outputs.keys(), universe=dataId.universe)
        self.prerequisites = _DatasetDict.fromDatasetTypes(
            task.prerequisites.keys(), universe=dataId.universe
        )

    __slots__ = ("task", "dataId", "inputs", "outputs", "prerequisites")

    def __repr__(self) -> str:
        return f"_QuantumScaffolding(taskDef={self.task.taskDef}, dataId={self.dataId}, ...)"

    task: _TaskScaffolding
    """Back-reference to the helper object for the `PipelineTask` this quantum
    represents an execution of.
    """

    dataId: DataCoordinate
    """Data ID for this quantum.
    """

    inputs: _DatasetDict
    """Nested dictionary containing `DatasetRef` inputs to this quantum.

    This is initialized to map each `DatasetType` to an empty dictionary at
    construction.  Those nested dictionaries are populated (with data IDs as
    keys) with unresolved `DatasetRef` instances in
    `_PipelineScaffolding.connectDataIds`.
    """

    outputs: _DatasetDict
    """Nested dictionary containing `DatasetRef` outputs this quantum.
    """

    prerequisites: _DatasetDict
    """Nested dictionary containing `DatasetRef` prerequisite inputs to this
    quantum.
    """

    def makeQuantum(self, datastore_records: Optional[Mapping[str, DatastoreRecordData]] = None) -> Quantum:
        """Transform the scaffolding object into a true `Quantum` instance.

        Parameters
        ----------
        datastore_records : `dict` [ `str`, `DatastoreRecordData` ], optional
            If not `None` then fill datastore records in each generated Quantum
            using the records from this structure.

        Returns
        -------
        quantum : `Quantum`
            An actual `Quantum` instance.
        """
        allInputs = self.inputs.unpackMultiRefs()
        allInputs.update(self.prerequisites.unpackMultiRefs())
        # Give the task's Connections class an opportunity to remove some
        # inputs, or complain if they are unacceptable.
        # This will raise if one of the check conditions is not met, which is
        # the intended behavior.
        # If it raises NotWorkFound, there is a bug in the QG algorithm
        # or the adjustQuantum is incorrectly trying to make a prerequisite
        # input behave like a regular input; adjustQuantum should only raise
        # NoWorkFound if a regular input is missing, and it shouldn't be
        # possible for us to have generated ``self`` if that's true.
        helper = AdjustQuantumHelper(inputs=allInputs, outputs=self.outputs.unpackMultiRefs())
        helper.adjust_in_place(self.task.taskDef.connections, self.task.taskDef.label, self.dataId)
        initInputs = self.task.initInputs.unpackSingleRefs()
        quantum_records: Optional[Mapping[str, DatastoreRecordData]] = None
        if datastore_records is not None:
            quantum_records = {}
            input_refs = list(itertools.chain.from_iterable(helper.inputs.values()))
            input_refs += list(initInputs.values())
            input_ids = set(ref.id for ref in input_refs if ref.id is not None)
            for datastore_name, records in datastore_records.items():
                matching_records = records.subset(input_ids)
                if matching_records is not None:
                    quantum_records[datastore_name] = matching_records
        return Quantum(
            taskName=self.task.taskDef.taskName,
            taskClass=self.task.taskDef.taskClass,
            dataId=self.dataId,
            initInputs=initInputs,
            inputs=helper.inputs,
            outputs=helper.outputs,
            datastore_records=quantum_records,
        )


@dataclass
class _TaskScaffolding:
    """Helper class aggregating information about a `PipelineTask`, used when
    constructing a `QuantumGraph`.

    See `_PipelineScaffolding` for a top-down description of the full
    scaffolding data structure.

    Parameters
    ----------
    taskDef : `TaskDef`
        Data structure that identifies the task class and its config.
    parent : `_PipelineScaffolding`
        The parent data structure that will hold the instance being
        constructed.
    datasetTypes : `TaskDatasetTypes`
        Data structure that categorizes the dataset types used by this task.
    """

    def __init__(self, taskDef: TaskDef, parent: _PipelineScaffolding, datasetTypes: TaskDatasetTypes):
        universe = parent.dimensions.universe
        self.taskDef = taskDef
        self.dimensions = DimensionGraph(universe, names=taskDef.connections.dimensions)
        assert self.dimensions.issubset(parent.dimensions)
        # Initialize _DatasetDicts as subsets of the one or two
        # corresponding dicts in the parent _PipelineScaffolding.
        self.initInputs = _DatasetDict.fromSubset(
            datasetTypes.initInputs, parent.initInputs, parent.initIntermediates
        )
        self.initOutputs = _DatasetDict.fromSubset(
            datasetTypes.initOutputs, parent.initIntermediates, parent.initOutputs
        )
        self.inputs = _DatasetDict.fromSubset(datasetTypes.inputs, parent.inputs, parent.intermediates)
        self.outputs = _DatasetDict.fromSubset(datasetTypes.outputs, parent.intermediates, parent.outputs)
        self.prerequisites = _DatasetDict.fromSubset(datasetTypes.prerequisites, parent.prerequisites)
        self.dataIds: Set[DataCoordinate] = set()
        self.quanta = {}

    def __repr__(self) -> str:
        # Default dataclass-injected __repr__ gets caught in an infinite loop
        # because of back-references.
        return f"_TaskScaffolding(taskDef={self.taskDef}, ...)"

    taskDef: TaskDef
    """Data structure that identifies the task class and its config
    (`TaskDef`).
    """

    dimensions: DimensionGraph
    """The dimensions of a single `Quantum` of this task (`DimensionGraph`).
    """

    initInputs: _DatasetDict
    """Dictionary containing information about datasets used to construct this
    task (`_DatasetDict`).
    """

    initOutputs: _DatasetDict
    """Dictionary containing information about datasets produced as a
    side-effect of constructing this task (`_DatasetDict`).
    """

    inputs: _DatasetDict
    """Dictionary containing information about datasets used as regular,
    graph-constraining inputs to this task (`_DatasetDict`).
    """

    outputs: _DatasetDict
    """Dictionary containing information about datasets produced by this task
    (`_DatasetDict`).
    """

    prerequisites: _DatasetDict
    """Dictionary containing information about input datasets that must be
    present in the repository before any Pipeline containing this task is run
    (`_DatasetDict`).
    """

    quanta: Dict[DataCoordinate, _QuantumScaffolding]
    """Dictionary mapping data ID to a scaffolding object for the Quantum of
    this task with that data ID.
    """

    def makeQuantumSet(
        self,
        unresolvedRefs: Optional[Set[DatasetRef]] = None,
        datastore_records: Optional[Mapping[str, DatastoreRecordData]] = None,
    ) -> Set[Quantum]:
        """Create a `set` of `Quantum` from the information in ``self``.

        Parameters
        ----------
        unresolvedRefs : `set` [ `DatasetRef` ], optional
            Input dataset refs that have not been found.
        datastore_records : `dict`


        Returns
        -------
        nodes : `set` of `Quantum`
            The `Quantum` elements corresponding to this task.
        """
        if unresolvedRefs is None:
            unresolvedRefs = set()
        outputs = set()
        for q in self.quanta.values():
            try:
                tmpQuanta = q.makeQuantum(datastore_records)
                outputs.add(tmpQuanta)
            except (NoWorkFound, FileNotFoundError) as exc:
                refs = itertools.chain.from_iterable(self.inputs.unpackMultiRefs().values())
                if unresolvedRefs.intersection(refs):
                    # This means it is a node that is Known to be pruned
                    # later and should be left in even though some follow up
                    # queries fail. This allows the pruning to start from this
                    # quantum with known issues, and prune other nodes it
                    # touches
                    inputs = q.inputs.unpackMultiRefs()
                    inputs.update(q.prerequisites.unpackMultiRefs())
                    tmpQuantum = Quantum(
                        taskName=q.task.taskDef.taskName,
                        taskClass=q.task.taskDef.taskClass,
                        dataId=q.dataId,
                        initInputs=q.task.initInputs.unpackSingleRefs(),
                        inputs=inputs,
                        outputs=q.outputs.unpackMultiRefs(),
                    )
                    outputs.add(tmpQuantum)
                else:
                    raise exc
        return outputs


@dataclass
class _PipelineScaffolding:
    """A helper data structure that organizes the information involved in
    constructing a `QuantumGraph` for a `Pipeline`.

    Parameters
    ----------
    pipeline : `Pipeline` or `Iterable` [ `TaskDef` ]
        Sequence of tasks from which a graph is to be constructed.  Must
        have nested task classes already imported.
    universe : `DimensionUniverse`
        Universe of all possible dimensions.

    Notes
    -----
    The scaffolding data structure contains nested data structures for both
    tasks (`_TaskScaffolding`) and datasets (`_DatasetDict`).  The dataset
    data structures are shared between the pipeline-level structure (which
    aggregates all datasets and categorizes them from the perspective of the
    complete pipeline) and the individual tasks that use them as inputs and
    outputs.

    `QuantumGraph` construction proceeds in four steps, with each corresponding
    to a different `_PipelineScaffolding` method:

    1. When `_PipelineScaffolding` is constructed, we extract and categorize
       the DatasetTypes used by the pipeline (delegating to
       `PipelineDatasetTypes.fromPipeline`), then use these to construct the
       nested `_TaskScaffolding` and `_DatasetDict` objects.

    2. In `connectDataIds`, we construct and run the "Big Join Query", which
       returns related tuples of all dimensions used to identify any regular
       input, output, and intermediate datasets (not prerequisites).  We then
       iterate over these tuples of related dimensions, identifying the subsets
       that correspond to distinct data IDs for each task and dataset type,
       and then create `_QuantumScaffolding` objects.

    3. In `resolveDatasetRefs`, we run follow-up queries against all of the
       dataset data IDs previously identified, transforming unresolved
       DatasetRefs into resolved DatasetRefs where appropriate.  We then look
       up prerequisite datasets for all quanta.

    4. In `makeQuantumGraph`, we construct a `QuantumGraph` from the lists of
       per-task `_QuantumScaffolding` objects.
    """

    def __init__(self, pipeline: Union[Pipeline, Iterable[TaskDef]], *, registry: Registry):
        _LOG.debug("Initializing data structures for QuantumGraph generation.")
        self.tasks = []
        # Aggregate and categorize the DatasetTypes in the Pipeline.
        datasetTypes = PipelineDatasetTypes.fromPipeline(pipeline, registry=registry)
        # Construct dictionaries that map those DatasetTypes to structures
        # that will (later) hold addiitonal information about them.
        for attr in (
            "initInputs",
            "initIntermediates",
            "initOutputs",
            "inputs",
            "intermediates",
            "outputs",
            "prerequisites",
        ):
            setattr(
                self,
                attr,
                _DatasetDict.fromDatasetTypes(getattr(datasetTypes, attr), universe=registry.dimensions),
            )
        self.defaultDatasetQueryConstraints = datasetTypes.queryConstraints
        # Aggregate all dimensions for all non-init, non-prerequisite
        # DatasetTypes.  These are the ones we'll include in the big join
        # query.
        self.dimensions = self.inputs.dimensions.union(self.intermediates.dimensions, self.outputs.dimensions)
        # Construct scaffolding nodes for each Task, and add backreferences
        # to the Task from each DatasetScaffolding node.
        # Note that there's only one scaffolding node for each DatasetType,
        # shared by _PipelineScaffolding and all _TaskScaffoldings that
        # reference it.
        if isinstance(pipeline, Pipeline):
            pipeline = pipeline.toExpandedPipeline()
        self.tasks = [
            _TaskScaffolding(taskDef=taskDef, parent=self, datasetTypes=taskDatasetTypes)
            for taskDef, taskDatasetTypes in zip(pipeline, datasetTypes.byTask.values())
        ]

    def __repr__(self) -> str:
        # Default dataclass-injected __repr__ gets caught in an infinite loop
        # because of back-references.
        return f"_PipelineScaffolding(tasks={self.tasks}, ...)"

    tasks: List[_TaskScaffolding]
    """Scaffolding data structures for each task in the pipeline
    (`list` of `_TaskScaffolding`).
    """

    initInputs: _DatasetDict
    """Datasets consumed but not produced when constructing the tasks in this
    pipeline (`_DatasetDict`).
    """

    initIntermediates: _DatasetDict
    """Datasets that are both consumed and produced when constructing the tasks
    in this pipeline (`_DatasetDict`).
    """

    initOutputs: _DatasetDict
    """Datasets produced but not consumed when constructing the tasks in this
    pipeline (`_DatasetDict`).
    """

    inputs: _DatasetDict
    """Datasets that are consumed but not produced when running this pipeline
    (`_DatasetDict`).
    """

    intermediates: _DatasetDict
    """Datasets that are both produced and consumed when running this pipeline
    (`_DatasetDict`).
    """

    outputs: _DatasetDict
    """Datasets produced but not consumed when when running this pipeline
    (`_DatasetDict`).
    """

    prerequisites: _DatasetDict
    """Datasets that are consumed when running this pipeline and looked up
    per-Quantum when generating the graph (`_DatasetDict`).
    """

    defaultDatasetQueryConstraints: NamedValueSet[DatasetType]
    """Datasets that should be used as constraints in the initial query,
    according to tasks (`NamedValueSet`).
    """

    dimensions: DimensionGraph
    """All dimensions used by any regular input, intermediate, or output
    (not prerequisite) dataset; the set of dimension used in the "Big Join
    Query" (`DimensionGraph`).

    This is required to be a superset of all task quantum dimensions.
    """

    @contextmanager
    def connectDataIds(
        self,
        registry: Registry,
        collections: Any,
        userQuery: Optional[str],
        externalDataId: DataCoordinate,
        datasetQueryConstraint: DatasetQueryConstraintVariant = DatasetQueryConstraintVariant.ALL,
    ) -> Iterator[DataCoordinateQueryResults]:
        """Query for the data IDs that connect nodes in the `QuantumGraph`.

        This method populates `_TaskScaffolding.dataIds` and
        `_DatasetScaffolding.dataIds` (except for those in `prerequisites`).

        Parameters
        ----------
        registry : `lsst.daf.butler.Registry`
            Registry for the data repository; used for all data ID queries.
        collections
            Expressions representing the collections to search for input
            datasets.  May be any of the types accepted by
            `lsst.daf.butler.CollectionSearch.fromExpression`.
        userQuery : `str` or `None`
            User-provided expression to limit the data IDs processed.
        externalDataId : `DataCoordinate`
            Externally-provided data ID that should be used to restrict the
            results, just as if these constraints had been included via ``AND``
            in ``userQuery``.  This includes (at least) any instrument named
            in the pipeline definition.
        datasetQueryConstraint : `DatasetQueryConstraintVariant`, optional
            The query constraint variant that should be used to constraint the
            query based on dataset existance, defaults to
            `DatasetQueryConstraintVariant.ALL`.

        Returns
        -------
        commonDataIds : \
                `lsst.daf.butler.registry.queries.DataCoordinateQueryResults`
            An interface to a database temporary table containing all data IDs
            that will appear in this `QuantumGraph`.  Returned inside a
            context manager, which will drop the temporary table at the end of
            the `with` block in which this method is called.
        """
        _LOG.debug("Building query for data IDs.")
        # Initialization datasets always have empty data IDs.
        emptyDataId = DataCoordinate.makeEmpty(registry.dimensions)
        for datasetType, refs in itertools.chain(
            self.initInputs.items(), self.initIntermediates.items(), self.initOutputs.items()
        ):
            refs[emptyDataId] = DatasetRef(datasetType, emptyDataId)
        # Run one big query for the data IDs for task dimensions and regular
        # inputs and outputs.  We limit the query to only dimensions that are
        # associated with the input dataset types, but don't (yet) try to
        # obtain the dataset_ids for those inputs.
        _LOG.debug(
            "Submitting data ID query over dimensions %s and materializing results.",
            list(self.dimensions.names),
        )
        queryArgs: Dict[str, Any] = {
            "dimensions": self.dimensions,
            "where": userQuery,
            "dataId": externalDataId,
        }
        if datasetQueryConstraint == DatasetQueryConstraintVariant.ALL:
            _LOG.debug(
                "Constraining graph query using default of %s.",
                list(self.defaultDatasetQueryConstraints.names),
            )
            queryArgs["datasets"] = list(self.defaultDatasetQueryConstraints)
            queryArgs["collections"] = collections
        elif datasetQueryConstraint == DatasetQueryConstraintVariant.OFF:
            _LOG.debug("Not using dataset existence to constrain query.")
        elif datasetQueryConstraint == DatasetQueryConstraintVariant.LIST:
            constraint = set(datasetQueryConstraint)
            inputs = {k.name: k for k in self.inputs.keys()}
            if remainder := constraint.difference(inputs.keys()):
                raise ValueError(
                    f"{remainder} dataset type(s) specified as a graph constraint, but"
                    f" do not appear as an input to the specified pipeline: {inputs.keys()}"
                )
            _LOG.debug(f"Constraining graph query using {constraint}")
            queryArgs["datasets"] = [typ for name, typ in inputs.items() if name in constraint]
            queryArgs["collections"] = collections
        else:
            raise ValueError(
                f"Unable to handle type {datasetQueryConstraint} given as datasetQueryConstraint."
            )

        with registry.queryDataIds(**queryArgs).materialize() as commonDataIds:
            _LOG.debug("Expanding data IDs.")
            commonDataIds = commonDataIds.expanded()
            _LOG.debug("Iterating over query results to associate quanta with datasets.")
            # Iterate over query results, populating data IDs for datasets and
            # quanta and then connecting them to each other.
            n = -1
            for n, commonDataId in enumerate(commonDataIds):
                # Create DatasetRefs for all DatasetTypes from this result row,
                # noting that we might have created some already.
                # We remember both those that already existed and those that we
                # create now.
                refsForRow = {}
                dataIdCacheForRow: Dict[DimensionGraph, DataCoordinate] = {}
                for datasetType, refs in itertools.chain(
                    self.inputs.items(), self.intermediates.items(), self.outputs.items()
                ):
                    datasetDataId: Optional[DataCoordinate]
                    if (datasetDataId := dataIdCacheForRow.get(datasetType.dimensions)) is None:
                        datasetDataId = commonDataId.subset(datasetType.dimensions)
                        dataIdCacheForRow[datasetType.dimensions] = datasetDataId
                    ref = refs.get(datasetDataId)
                    if ref is None:
                        ref = DatasetRef(datasetType, datasetDataId)
                        refs[datasetDataId] = ref
                    refsForRow[datasetType.name] = ref
                # Create _QuantumScaffolding objects for all tasks from this
                # result row, noting that we might have created some already.
                for task in self.tasks:
                    quantumDataId = commonDataId.subset(task.dimensions)
                    quantum = task.quanta.get(quantumDataId)
                    if quantum is None:
                        quantum = _QuantumScaffolding(task=task, dataId=quantumDataId)
                        task.quanta[quantumDataId] = quantum
                    # Whether this is a new quantum or an existing one, we can
                    # now associate the DatasetRefs for this row with it.  The
                    # fact that a Quantum data ID and a dataset data ID both
                    # came from the same result row is what tells us they
                    # should be associated.
                    # Many of these associates will be duplicates (because
                    # another query row that differed from this one only in
                    # irrelevant dimensions already added them), and we use
                    # sets to skip.
                    for datasetType in task.inputs:
                        ref = refsForRow[datasetType.name]
                        quantum.inputs[datasetType.name][ref.dataId] = ref
                    for datasetType in task.outputs:
                        ref = refsForRow[datasetType.name]
                        quantum.outputs[datasetType.name][ref.dataId] = ref
            if n < 0:
                emptiness_explained = False
                for message in commonDataIds.explain_no_results():
                    _LOG.warning(message)
                    emptiness_explained = True
                if not emptiness_explained:
                    _LOG.warning(
                        "To reproduce this query for debugging purposes, run "
                        "Registry.queryDataIds with these arguments:"
                    )
                    # We could just repr() the queryArgs dict to get something
                    # the user could make sense of, but it's friendlier to
                    # put these args in an easier-to-construct equivalent form
                    # so they can read it more easily and copy and paste into
                    # a Python terminal.
                    _LOG.warning("  dimensions=%s,", list(queryArgs["dimensions"].names))
                    _LOG.warning("  dataId=%s,", queryArgs["dataId"].byName())
                    if queryArgs["where"]:
                        _LOG.warning("  where=%s,", repr(queryArgs["where"]))
                    if "datasets" in queryArgs:
                        _LOG.warning("  datasets=%s,", [t.name for t in queryArgs["datasets"]])
                    if "collections" in queryArgs:
                        _LOG.warning("  collections=%s,", list(queryArgs["collections"]))
            _LOG.debug("Finished processing %d rows from data ID query.", n)
            yield commonDataIds

    def resolveDatasetRefs(
        self,
        registry: Registry,
        collections: Any,
        run: Optional[str],
        commonDataIds: DataCoordinateQueryResults,
        *,
        skipExistingIn: Any = None,
        clobberOutputs: bool = True,
        constrainedByAllDatasets: bool = True,
    ) -> None:
        """Perform follow up queries for each dataset data ID produced in
        `fillDataIds`.

        This method populates `_DatasetScaffolding.refs` (except for those in
        `prerequisites`).

        Parameters
        ----------
        registry : `lsst.daf.butler.Registry`
            Registry for the data repository; used for all data ID queries.
        collections
            Expressions representing the collections to search for input
            datasets.  May be any of the types accepted by
            `lsst.daf.butler.CollectionSearch.fromExpression`.
        run : `str`, optional
            Name of the `~lsst.daf.butler.CollectionType.RUN` collection for
            output datasets, if it already exists.
        commonDataIds : \
                `lsst.daf.butler.registry.queries.DataCoordinateQueryResults`
            Result of a previous call to `connectDataIds`.
        skipExistingIn
            Expressions representing the collections to search for existing
            output datasets that should be skipped.  May be any of the types
            accepted by `lsst.daf.butler.CollectionSearch.fromExpression`.
            `None` or empty string/sequence disables skipping.
        clobberOutputs : `bool`, optional
            If `True` (default), allow quanta to created even if outputs exist;
            this requires the same behavior behavior to be enabled when
            executing.  If ``skipExistingIn`` is not `None`, completed quanta
            (those with metadata, or all outputs if there is no metadata
            dataset configured) will be skipped rather than clobbered.
        constrainedByAllDatasets : `bool`, optional
            Indicates if the commonDataIds were generated with a constraint on
            all dataset types.

        Raises
        ------
        OutputExistsError
            Raised if an output dataset already exists in the output run
            and ``skipExistingIn`` does not include output run, or if only
            some outputs are present and ``clobberOutputs`` is `False`.
        """
        skipCollections: Optional[CollectionSearch] = None
        skipExistingInRun = False
        if skipExistingIn:
            skipCollections = CollectionSearch.fromExpression(skipExistingIn)
            if run:
                # as optimization check in the explicit list of names first
                skipExistingInRun = run in skipCollections.explicitNames()
                if not skipExistingInRun:
                    # need to flatten it and check again
                    skipExistingInRun = run in registry.queryCollections(
                        skipExistingIn,
                        collectionTypes=CollectionType.RUN,
                    )

        # Updating constrainedByAllDatasets here is not ideal, but we have a
        # few different code paths that each transfer different pieces of
        # information about what dataset query constraints were applied here,
        # and none of them has the complete picture until we get here.  We're
        # long overdue for a QG generation rewrite that will make this go away
        # entirely anyway.
        constrainedByAllDatasets = (
            constrainedByAllDatasets and self.defaultDatasetQueryConstraints == self.inputs.keys()
        )

        # Look up [init] intermediate and output datasets in the output
        # collection, if there is an output collection.
        if run is not None or skipCollections is not None:
            for datasetType, refs in itertools.chain(
                self.initIntermediates.items(),
                self.initOutputs.items(),
                self.intermediates.items(),
                self.outputs.items(),
            ):
                _LOG.debug(
                    "Resolving %d datasets for intermediate and/or output dataset %s.",
                    len(refs),
                    datasetType.name,
                )
                isInit = datasetType in self.initIntermediates or datasetType in self.initOutputs
                subset = commonDataIds.subset(datasetType.dimensions, unique=True)

                # look at RUN collection first
                if run is not None:
                    resolvedRefQueryResults = subset.findDatasets(
                        datasetType, collections=run, findFirst=True
                    )
                    for resolvedRef in resolvedRefQueryResults:
                        # TODO: we could easily support per-DatasetType
                        # skipExisting and I could imagine that being useful -
                        # it's probably required in order to support writing
                        # initOutputs before QuantumGraph generation.
                        assert resolvedRef.dataId in refs
                        if not (skipExistingInRun or isInit or clobberOutputs):
                            raise OutputExistsError(
                                f"Output dataset {datasetType.name} already exists in "
                                f"output RUN collection '{run}' with data ID"
                                f" {resolvedRef.dataId}."
                            )

                # And check skipExistingIn too, if RUN collection is in
                # it is handled above
                if skipCollections is not None:
                    resolvedRefQueryResults = subset.findDatasets(
                        datasetType, collections=skipCollections, findFirst=True
                    )
                    for resolvedRef in resolvedRefQueryResults:
                        assert resolvedRef.dataId in refs
                        refs[resolvedRef.dataId] = resolvedRef

        # Look up input and initInput datasets in the input collection(s).
        # container to accumulate unfound refs, if the common dataIs were not
        # constrained on dataset type existence.
        self.unfoundRefs = set()
        for datasetType, refs in itertools.chain(self.initInputs.items(), self.inputs.items()):
            _LOG.debug("Resolving %d datasets for input dataset %s.", len(refs), datasetType.name)
            resolvedRefQueryResults = commonDataIds.subset(datasetType.dimensions, unique=True).findDatasets(
                datasetType, collections=collections, findFirst=True
            )
            dataIdsNotFoundYet = set(refs.keys())
            for resolvedRef in resolvedRefQueryResults:
                dataIdsNotFoundYet.discard(resolvedRef.dataId)
                refs[resolvedRef.dataId] = resolvedRef
            if dataIdsNotFoundYet:
                if constrainedByAllDatasets:
                    raise RuntimeError(
                        f"{len(dataIdsNotFoundYet)} dataset(s) of type "
                        f"'{datasetType.name}' was/were present in a previous "
                        f"query, but could not be found now."
                        f"This is either a logic bug in QuantumGraph generation "
                        f"or the input collections have been modified since "
                        f"QuantumGraph generation began."
                    )
                elif not datasetType.dimensions:
                    raise RuntimeError(
                        f"Dataset {datasetType.name!r} (with no dimensions) could not be found in "
                        f"collections {collections}."
                    )
                else:
                    # if the common dataIds were not constrained using all the
                    # input dataset types, it is possible that some data ids
                    # found dont correspond to existing dataset types and they
                    # will be un-resolved. Mark these for later pruning from
                    # the quantum graph.
                    for k in dataIdsNotFoundYet:
                        self.unfoundRefs.add(refs[k])

        # Copy the resolved DatasetRefs to the _QuantumScaffolding objects,
        # replacing the unresolved refs there, and then look up prerequisites.
        for task in self.tasks:
            _LOG.debug(
                "Applying resolutions and finding prerequisites for %d quanta of task with label '%s'.",
                len(task.quanta),
                task.taskDef.label,
            )
            # The way iterConnections is designed makes it impossible to
            # annotate precisely enough to satisfy MyPy here.
            lookupFunctions = {
                c.name: c.lookupFunction  # type: ignore
                for c in iterConnections(task.taskDef.connections, "prerequisiteInputs")
                if c.lookupFunction is not None  # type: ignore
            }
            dataIdsFailed = []
            dataIdsSucceeded = []
            for quantum in task.quanta.values():
                # Process outputs datasets only if skipExistingIn is not None
                # or there is a run to look for outputs in and clobberOutputs
                # is True. Note that if skipExistingIn is None, any output
                # datasets that already exist would have already caused an
                # exception to be raised.  We never update the DatasetRefs in
                # the quantum because those should never be resolved.
                if skipCollections is not None or (run is not None and clobberOutputs):
                    resolvedRefs = []
                    unresolvedRefs = []
                    haveMetadata = False
                    for datasetType, originalRefs in quantum.outputs.items():
                        for ref in task.outputs.extract(datasetType, originalRefs.keys()):
                            if ref.id is not None:
                                resolvedRefs.append(ref)
                                if datasetType.name == task.taskDef.metadataDatasetName:
                                    haveMetadata = True
                            else:
                                unresolvedRefs.append(ref)
                    if resolvedRefs:
                        if haveMetadata or not unresolvedRefs:
                            dataIdsSucceeded.append(quantum.dataId)
                            if skipCollections is not None:
                                continue
                        else:
                            dataIdsFailed.append(quantum.dataId)
                            if not clobberOutputs:
                                raise OutputExistsError(
                                    f"Quantum {quantum.dataId} of task with label "
                                    f"'{quantum.task.taskDef.label}' has some outputs that exist "
                                    f"({resolvedRefs}) "
                                    f"and others that don't ({unresolvedRefs}), with no metadata output, "
                                    "and clobbering outputs was not enabled."
                                )
                # Update the input DatasetRefs to the resolved ones we already
                # searched for.
                for datasetType, input_refs in quantum.inputs.items():
                    for ref in task.inputs.extract(datasetType, input_refs.keys()):
                        input_refs[ref.dataId] = ref
                # Look up prerequisite datasets in the input collection(s).
                # These may have dimensions that extend beyond those we queried
                # for originally, because we want to permit those data ID
                # values to differ across quanta and dataset types.
                for datasetType in task.prerequisites:
                    lookupFunction = lookupFunctions.get(datasetType.name)
                    if lookupFunction is not None:
                        # PipelineTask has provided its own function to do the
                        # lookup.  This always takes precedence.
                        prereq_refs = list(lookupFunction(datasetType, registry, quantum.dataId, collections))
                    elif (
                        datasetType.isCalibration()
                        and datasetType.dimensions <= quantum.dataId.graph
                        and quantum.dataId.graph.temporal
                    ):
                        # This is a master calibration lookup, which we have to
                        # handle specially because the query system can't do a
                        # temporal join on a non-dimension-based timespan yet.
                        timespan = quantum.dataId.timespan
                        try:
                            prereq_refs = [
                                registry.findDataset(
                                    datasetType, quantum.dataId, collections=collections, timespan=timespan
                                )
                            ]
                        except KeyError:
                            # This dataset type is not present in the registry,
                            # which just means there are no datasets here.
                            prereq_refs = []
                    else:
                        # Most general case.
                        prereq_refs = list(
                            registry.queryDatasets(
                                datasetType, collections=collections, dataId=quantum.dataId, findFirst=True
                            ).expanded()
                        )
                    quantum.prerequisites[datasetType].update(
                        {ref.dataId: ref for ref in prereq_refs if ref is not None}
                    )
            # Actually remove any quanta that we decided to skip above.
            if dataIdsSucceeded:
                if skipCollections is not None:
                    _LOG.debug(
                        "Pruning successful %d quanta for task with label '%s' because all of their "
                        "outputs exist or metadata was written successfully.",
                        len(dataIdsSucceeded),
                        task.taskDef.label,
                    )
                    for dataId in dataIdsSucceeded:
                        del task.quanta[dataId]
                elif clobberOutputs:
                    _LOG.info(
                        "Found %d successful quanta for task with label '%s' "
                        "that will need to be clobbered during execution.",
                        len(dataIdsSucceeded),
                        task.taskDef.label,
                    )
                else:
                    raise AssertionError("OutputExistsError should have already been raised.")
            if dataIdsFailed:
                if clobberOutputs:
                    _LOG.info(
                        "Found %d failed/incomplete quanta for task with label '%s' "
                        "that will need to be clobbered during execution.",
                        len(dataIdsFailed),
                        task.taskDef.label,
                    )
                else:
                    raise AssertionError("OutputExistsError should have already been raised.")

    def makeQuantumGraph(
        self, metadata: Optional[Mapping[str, Any]] = None, datastore: Optional[Datastore] = None
    ) -> QuantumGraph:
        """Create a `QuantumGraph` from the quanta already present in
        the scaffolding data structure.

        Parameters
        ---------
        metadata : Optional Mapping of `str` to primitives
            This is an optional parameter of extra data to carry with the
            graph.  Entries in this mapping should be able to be serialized in
            JSON.
        datastore : `Datastore`, optional
            If not `None` then fill datastore records in each generated
            Quantum.

        Returns
        -------
        graph : `QuantumGraph`
            The full `QuantumGraph`.
        """

        def _make_refs(dataset_dict: _DatasetDict) -> Iterable[DatasetRef]:
            """Extract all DatasetRefs from the dictionaries"""
            for ref_dict in dataset_dict.values():
                yield from ref_dict.values()

        datastore_records: Optional[Mapping[str, DatastoreRecordData]] = None
        if datastore is not None:
            datastore_records = datastore.export_records(
                itertools.chain(
                    _make_refs(self.inputs), _make_refs(self.initInputs), _make_refs(self.prerequisites)
                )
            )

        graphInput: Dict[TaskDef, Set[Quantum]] = {}
        for task in self.tasks:
            qset = task.makeQuantumSet(unresolvedRefs=self.unfoundRefs, datastore_records=datastore_records)
            graphInput[task.taskDef] = qset

        graph = QuantumGraph(
            graphInput, metadata=metadata, pruneRefs=self.unfoundRefs, universe=self.dimensions.universe
        )
        return graph


# ------------------------
#  Exported definitions --
# ------------------------


class GraphBuilderError(Exception):
    """Base class for exceptions generated by graph builder."""

    pass


class OutputExistsError(GraphBuilderError):
    """Exception generated when output datasets already exist."""

    pass


class PrerequisiteMissingError(GraphBuilderError):
    """Exception generated when a prerequisite dataset does not exist."""

    pass


class GraphBuilder:
    """GraphBuilder class is responsible for building task execution graph from
    a Pipeline.

    Parameters
    ----------
    registry : `~lsst.daf.butler.Registry`
        Data butler instance.
    skipExistingIn
        Expressions representing the collections to search for existing
        output datasets that should be skipped.  May be any of the types
        accepted by `lsst.daf.butler.CollectionSearch.fromExpression`.
    clobberOutputs : `bool`, optional
        If `True` (default), allow quanta to created even if partial outputs
        exist; this requires the same behavior behavior to be enabled when
        executing.
    datastore : `Datastore`, optional
        If not `None` then fill datastore records in each generated Quantum.
    """

    def __init__(
        self,
        registry: Registry,
        skipExistingIn: Any = None,
        clobberOutputs: bool = True,
        datastore: Optional[Datastore] = None,
    ):
        self.registry = registry
        self.dimensions = registry.dimensions
        self.skipExistingIn = skipExistingIn
        self.clobberOutputs = clobberOutputs
        self.datastore = datastore

    def makeGraph(
        self,
        pipeline: Union[Pipeline, Iterable[TaskDef]],
        collections: Any,
        run: Optional[str],
        userQuery: Optional[str],
        datasetQueryConstraint: DatasetQueryConstraintVariant = DatasetQueryConstraintVariant.ALL,
        metadata: Optional[Mapping[str, Any]] = None,
    ) -> QuantumGraph:
        """Create execution graph for a pipeline.

        Parameters
        ----------
        pipeline : `Pipeline` or `Iterable` [ `TaskDef` ]
            Pipeline definition, task names/classes and their configs.
        collections
            Expressions representing the collections to search for input
            datasets.  May be any of the types accepted by
            `lsst.daf.butler.CollectionSearch.fromExpression`.
        run : `str`, optional
            Name of the `~lsst.daf.butler.CollectionType.RUN` collection for
            output datasets, if it already exists.
        userQuery : `str`
            String which defines user-defined selection for registry, should be
            empty or `None` if there is no restrictions on data selection.
        datasetQueryConstraint : `DatasetQueryConstraintVariant`, optional
            The query constraint variant that should be used to constraint the
            query based on dataset existance, defaults to
            `DatasetQueryConstraintVariant.ALL`.
        metadata : Optional Mapping of `str` to primitives
            This is an optional parameter of extra data to carry with the
            graph.  Entries in this mapping should be able to be serialized in
            JSON.

        Returns
        -------
        graph : `QuantumGraph`

        Raises
        ------
        UserExpressionError
            Raised when user expression cannot be parsed.
        OutputExistsError
            Raised when output datasets already exist.
        Exception
            Other exceptions types may be raised by underlying registry
            classes.
        """
        scaffolding = _PipelineScaffolding(pipeline, registry=self.registry)
        if not collections and (scaffolding.initInputs or scaffolding.inputs or scaffolding.prerequisites):
            raise ValueError("Pipeline requires input datasets but no input collections provided.")
        instrument_class: Optional[Any] = None
        if isinstance(pipeline, Pipeline):
            instrument_class_name = pipeline.getInstrument()
            if instrument_class_name is not None:
                instrument_class = doImportType(instrument_class_name)
            pipeline = list(pipeline.toExpandedPipeline())
        if instrument_class is not None:
            dataId = DataCoordinate.standardize(
                instrument=instrument_class.getName(), universe=self.registry.dimensions
            )
        else:
            dataId = DataCoordinate.makeEmpty(self.registry.dimensions)
        with scaffolding.connectDataIds(
            self.registry, collections, userQuery, dataId, datasetQueryConstraint
        ) as commonDataIds:
            condition = datasetQueryConstraint == DatasetQueryConstraintVariant.ALL
            scaffolding.resolveDatasetRefs(
                self.registry,
                collections,
                run,
                commonDataIds,
                skipExistingIn=self.skipExistingIn,
                clobberOutputs=self.clobberOutputs,
                constrainedByAllDatasets=condition,
            )
        return scaffolding.makeQuantumGraph(metadata=metadata, datastore=self.datastore)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Instrument",)

import datetime
import os.path
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Optional, Sequence, Type, Union

from lsst.daf.butler import DataId, Formatter
from lsst.daf.butler.registry import DataIdError
from lsst.utils import doImportType

if TYPE_CHECKING:
    from lsst.daf.butler import Registry
    from lsst.pex.config import Config


class Instrument(metaclass=ABCMeta):
    """Base class for instrument-specific logic for the Gen3 Butler.

    Parameters
    ----------
    collection_prefix : `str`, optional
        Prefix for collection names to use instead of the intrument's own name.
        This is primarily for use in simulated-data repositories, where the
        instrument name may not be necessary and/or sufficient to distinguish
        between collections.

    Notes
    -----
    Concrete instrument subclasses must have the same construction signature as
    the base class.
    """

    configPaths: Sequence[str] = ()
    """Paths to config files to read for specific Tasks.

    The paths in this list should contain files of the form `task.py`, for
    each of the Tasks that requires special configuration.
    """

    policyName: Optional[str] = None
    """Instrument specific name to use when locating a policy or configuration
    file in the file system."""

    def __init__(self, collection_prefix: Optional[str] = None):
        if collection_prefix is None:
            collection_prefix = self.getName()
        self.collection_prefix = collection_prefix

    @classmethod
    @abstractmethod
    def getName(cls) -> str:
        """Return the short (dimension) name for this instrument.

        This is not (in general) the same as the class name - it's what is used
        as the value of the "instrument" field in data IDs, and is usually an
        abbreviation of the full name.
        """
        raise NotImplementedError()

    @abstractmethod
    def register(self, registry: Registry, *, update: bool = False) -> None:
        """Insert instrument, and other relevant records into `Registry`.

        Parameters
        ----------
        registry : `lsst.daf.butler.Registry`
            Registry client for the data repository to modify.
        update : `bool`, optional
            If `True` (`False` is default), update existing records if they
            differ from the new ones.

        Raises
        ------
        lsst.daf.butler.registry.ConflictingDefinitionError
            Raised if any existing record has the same key but a different
            definition as one being registered.

        Notes
        -----
        New records can always be added by calling this method multiple times,
        as long as no existing records have changed (if existing records have
        changed, ``update=True`` must be used).  Old records can never be
        removed by this method.

        Implementations should guarantee that registration is atomic (the
        registry should not be modified if any error occurs) and idempotent at
        the level of individual dimension entries; new detectors and filters
        should be added, but changes to any existing record should not be.
        This can generally be achieved via a block like

        .. code-block:: python

            with registry.transaction():
                registry.syncDimensionData("instrument", ...)
                registry.syncDimensionData("detector", ...)
                self.registerFilters(registry)

        """
        raise NotImplementedError()

    @staticmethod
    def fromName(name: str, registry: Registry, collection_prefix: Optional[str] = None) -> Instrument:
        """Given an instrument name and a butler registry, retrieve a
        corresponding instantiated instrument object.

        Parameters
        ----------
        name : `str`
            Name of the instrument (must match the return value of `getName`).
        registry : `lsst.daf.butler.Registry`
            Butler registry to query to find the information.
        collection_prefix : `str`, optional
            Prefix for collection names to use instead of the intrument's own
            name.  This is primarily for use in simulated-data repositories,
            where the instrument name may not be necessary and/or sufficient to
            distinguish between collections.

        Returns
        -------
        instrument : `Instrument`
            An instance of the relevant `Instrument`.

        Notes
        -----
        The instrument must be registered in the corresponding butler.

        Raises
        ------
        LookupError
            Raised if the instrument is not known to the supplied registry.
        ModuleNotFoundError
            Raised if the class could not be imported.  This could mean
            that the relevant obs package has not been setup.
        TypeError
            Raised if the class name retrieved is not a string or the imported
            symbol is not an `Instrument` subclass.
        """
        try:
            records = list(registry.queryDimensionRecords("instrument", instrument=name))
        except DataIdError:
            records = None
        if not records:
            raise LookupError(f"No registered instrument with name '{name}'.")
        cls_name = records[0].class_name
        if not isinstance(cls_name, str):
            raise TypeError(
                f"Unexpected class name retrieved from {name} instrument dimension (got {cls_name})"
            )
        instrument_cls: type = doImportType(cls_name)
        if not issubclass(instrument_cls, Instrument):
            raise TypeError(
                f"{instrument_cls!r}, obtained from importing {cls_name}, is not an Instrument subclass."
            )
        return instrument_cls(collection_prefix=collection_prefix)

    @staticmethod
    def from_string(
        name: str, registry: Optional[Registry] = None, collection_prefix: Optional[str] = None
    ) -> Instrument:
        """Return an instance from the short name or class name.

        If the instrument name is not qualified (does not contain a '.') and a
        butler registry is provided, this will attempt to load the instrument
        using `Instrument.fromName()`. Otherwise the instrument will be
        imported and instantiated.

        Parameters
        ----------
        name : `str`
            The name or fully-qualified class name of an instrument.
        registry : `lsst.daf.butler.Registry`, optional
            Butler registry to query to find information about the instrument,
            by default `None`.
        collection_prefix : `str`, optional
            Prefix for collection names to use instead of the intrument's own
            name. This is primarily for use in simulated-data repositories,
            where the instrument name may not be necessary and/or sufficient
            to distinguish between collections.

        Returns
        -------
        instrument : `Instrument`
            The instantiated instrument.

        Raises
        ------
        RuntimeError
            Raised if the instrument can not be imported, instantiated, or
            obtained from the registry.
        TypeError
            Raised if the instrument is not a subclass of
            `~lsst.pipe.base.Instrument`.

        See Also
        --------
        Instrument.fromName
        """
        if "." not in name and registry is not None:
            try:
                instr = Instrument.fromName(name, registry, collection_prefix=collection_prefix)
            except Exception as err:
                raise RuntimeError(
                    f"Could not get instrument from name: {name}. Failed with exception: {err}"
                ) from err
        else:
            try:
                instr_class = doImportType(name)
            except Exception as err:
                raise RuntimeError(
                    f"Could not import instrument: {name}. Failed with exception: {err}"
                ) from err
            instr = instr_class(collection_prefix=collection_prefix)
        if not isinstance(instr, Instrument):
            raise TypeError(f"{name} is not an Instrument subclass.")
        return instr

    @staticmethod
    def importAll(registry: Registry) -> None:
        """Import all the instruments known to this registry.

        This will ensure that all metadata translators have been registered.

        Parameters
        ----------
        registry : `lsst.daf.butler.Registry`
            Butler registry to query to find the information.

        Notes
        -----
        It is allowed for a particular instrument class to fail on import.
        This might simply indicate that a particular obs package has
        not been setup.
        """
        records = list(registry.queryDimensionRecords("instrument"))
        for record in records:
            cls = record.class_name
            try:
                doImportType(cls)
            except Exception:
                pass

    @abstractmethod
    def getRawFormatter(self, dataId: DataId) -> Type[Formatter]:
        """Return the Formatter class that should be used to read a particular
        raw file.

        Parameters
        ----------
        dataId : `DataId`
            Dimension-based ID for the raw file or files being ingested.

        Returns
        -------
        formatter : `lsst.daf.butler.Formatter` class
            Class to be used that reads the file into the correct
            Python object for the raw data.
        """
        raise NotImplementedError()

    def applyConfigOverrides(self, name: str, config: Config) -> None:
        """Apply instrument-specific overrides for a task config.

        Parameters
        ----------
        name : `str`
            Name of the object being configured; typically the _DefaultName
            of a Task.
        config : `lsst.pex.config.Config`
            Config instance to which overrides should be applied.
        """
        for root in self.configPaths:
            path = os.path.join(root, f"{name}.py")
            if os.path.exists(path):
                config.load(path)

    @staticmethod
    def formatCollectionTimestamp(timestamp: Union[str, datetime.datetime]) -> str:
        """Format a timestamp for use in a collection name.

        Parameters
        ----------
        timestamp : `str` or `datetime.datetime`
            Timestamp to format.  May be a date or datetime string in extended
            ISO format (assumed UTC), with or without a timezone specifier, a
            datetime string in basic ISO format with a timezone specifier, a
            naive `datetime.datetime` instance (assumed UTC) or a
            timezone-aware `datetime.datetime` instance (converted to UTC).
            This is intended to cover all forms that string ``CALIBDATE``
            metadata values have taken in the past, as well as the format this
            method itself writes out (to enable round-tripping).

        Returns
        -------
        formatted : `str`
            Standardized string form for the timestamp.
        """
        if isinstance(timestamp, str):
            if "-" in timestamp:
                # extended ISO format, with - and : delimiters
                timestamp = datetime.datetime.fromisoformat(timestamp)
            else:
                # basic ISO format, with no delimiters (what this method
                # returns)
                timestamp = datetime.datetime.strptime(timestamp, "%Y%m%dT%H%M%S%z")
        if not isinstance(timestamp, datetime.datetime):
            raise TypeError(f"Unexpected date/time object: {timestamp!r}.")
        if timestamp.tzinfo is not None:
            timestamp = timestamp.astimezone(datetime.timezone.utc)
        return f"{timestamp:%Y%m%dT%H%M%S}Z"

    @staticmethod
    def makeCollectionTimestamp() -> str:
        """Create a timestamp string for use in a collection name from the
        current time.

        Returns
        -------
        formatted : `str`
            Standardized string form of the current time.
        """
        return Instrument.formatCollectionTimestamp(datetime.datetime.now(tz=datetime.timezone.utc))

    def makeDefaultRawIngestRunName(self) -> str:
        """Make the default instrument-specific run collection string for raw
        data ingest.

        Returns
        -------
        coll : `str`
            Run collection name to be used as the default for ingestion of
            raws.
        """
        return self.makeCollectionName("raw", "all")

    def makeUnboundedCalibrationRunName(self, *labels: str) -> str:
        """Make a RUN collection name appropriate for inserting calibration
        datasets whose validity ranges are unbounded.

        Parameters
        ----------
        *labels : `str`
            Extra strings to be included in the base name, using the default
            delimiter for collection names.  Usually this is the name of the
            ticket on which the calibration collection is being created.

        Returns
        -------
        name : `str`
            Run collection name.
        """
        return self.makeCollectionName("calib", *labels, "unbounded")

    def makeCuratedCalibrationRunName(self, calibDate: str, *labels: str) -> str:
        """Make a RUN collection name appropriate for inserting curated
        calibration datasets with the given ``CALIBDATE`` metadata value.

        Parameters
        ----------
        calibDate : `str`
            The ``CALIBDATE`` metadata value.
        *labels : `str`
            Strings to be included in the collection name (before
            ``calibDate``, but after all other terms), using the default
            delimiter for collection names.  Usually this is the name of the
            ticket on which the calibration collection is being created.

        Returns
        -------
        name : `str`
            Run collection name.
        """
        return self.makeCollectionName("calib", *labels, "curated", self.formatCollectionTimestamp(calibDate))

    def makeCalibrationCollectionName(self, *labels: str) -> str:
        """Make a CALIBRATION collection name appropriate for associating
        calibration datasets with validity ranges.

        Parameters
        ----------
        *labels : `str`
            Strings to be appended to the base name, using the default
            delimiter for collection names.  Usually this is the name of the
            ticket on which the calibration collection is being created.

        Returns
        -------
        name : `str`
            Calibration collection name.
        """
        return self.makeCollectionName("calib", *labels)

    @staticmethod
    def makeRefCatCollectionName(*labels: str) -> str:
        """Return a global (not instrument-specific) name for a collection that
        holds reference catalogs.

        With no arguments, this returns the name of the collection that holds
        all reference catalogs (usually a ``CHAINED`` collection, at least in
        long-lived repos that may contain more than one reference catalog).

        Parameters
        ----------
        *labels : `str`
            Strings to be added to the global collection name, in order to
            define a collection name for one or more reference catalogs being
            ingested at the same time.

        Returns
        -------
        name : `str`
            Collection name.

        Notes
        -----
        This is a ``staticmethod``, not a ``classmethod``, because it should
        be the same for all instruments.
        """
        return "/".join(("refcats",) + labels)

    def makeUmbrellaCollectionName(self) -> str:
        """Return the name of the umbrella ``CHAINED`` collection for this
        instrument that combines all standard recommended input collections.

        This method should almost never be overridden by derived classes.

        Returns
        -------
        name : `str`
            Name for the umbrella collection.
        """
        return self.makeCollectionName("defaults")

    def makeCollectionName(self, *labels: str) -> str:
        """Get the instrument-specific collection string to use as derived
        from the supplied labels.

        Parameters
        ----------
        *labels : `str`
            Strings to be combined with the instrument name to form a
            collection name.

        Returns
        -------
        name : `str`
            Collection name to use that includes the instrument's recommended
            prefix.
        """
        return "/".join((self.collection_prefix,) + labels)



#
# LSST Data Management System
# Copyright 2008, 2009, 2010, 2011 LSST Corporation.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <http://www.lsstcorp.org/LegalNotices/>.
#
"""Utilities for measuring execution time.
"""
__all__ = ["logInfo", "timeMethod"]

import logging
from typing import Any, Callable

import lsst.utils.timer
from deprecated.sphinx import deprecated


@deprecated(
    reason="logInfo has been replaced by lsst.utils.timer.logInfo. Will be removed after v25.",
    version="v24",
    category=FutureWarning,
)
def logInfo(obj, prefix, logLevel=logging.DEBUG, metadata=None, logger=None):
    """Log timer information to ``obj.metadata`` and ``obj.log``.

    Parameters
    ----------
    obj : `lsst.pipe.base.Task`-type or `None`
        A `~lsst.pipe.base.Task` or any other object with these two attributes:

        - ``metadata`` an instance of `~lsst.pipe.base.TaskMetadata` (or other
          object with ``add(name, value)`` method).
        - ``log`` an instance of `logging.Logger` or subclass.

        If `None`, at least one of ``metadata`` or ``logger`` should be passed
        or this function will do nothing.
    prefix : `str`
        Name prefix, the resulting entries are ``CpuTime``, etc.. For example
        timeMethod uses ``prefix = Start`` when the method begins and
        ``prefix = End`` when the method ends.
    logLevel : `int`, optional
        Log level (an `logging` level constant, such as `logging.DEBUG`).
    metadata : `lsst.pipe.base.TaskMetadata`, optional
        Metadata object to write entries to, overriding ``obj.metadata``.
    logger : `logging.Logger`
        Log object to write entries to, overriding ``obj.log``.

    Notes
    -----
    Logged items include:

    - ``Utc``: UTC date in ISO format (only in metadata since log entries have
      timestamps).
    - ``CpuTime``: System + User CPU time (seconds). This should only be used
        in differential measurements; the time reference point is undefined.
    - ``MaxRss``: maximum resident set size.

    All logged resource information is only for the current process; child
    processes are excluded.
    """
    return lsst.utils.timer.logInfo(obj, prefix, logLevel=logLevel, metadata=metadata, logger=logger)


@deprecated(
    reason="timeMethod has been replaced by lsst.utils.timer.timeMethod. Will be removed after v25.",
    version="v24",
    category=FutureWarning,
)
def timeMethod(*args: Any, **kwargs: Any) -> Callable:
    """Decorator to measure duration of a method.

    Notes
    -----
    This is a just a forwarding method for `lsst.utils.timer.timeMethod`. For
    documentation look at `lsst.utils.timer.timeMethod`.
    """
    return lsst.utils.timer.timeMethod(*args, **kwargs)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Unit tests for `lsst.pipe.base.tests`, a library for testing
PipelineTask subclasses.
"""

import shutil
import tempfile
import unittest

import lsst.daf.butler
import lsst.daf.butler.tests as butlerTests
import lsst.pex.config
import lsst.utils.tests
from lsst.pipe.base import PipelineTask, PipelineTaskConfig, PipelineTaskConnections, Struct, connectionTypes
from lsst.pipe.base.testUtils import (
    assertValidInitOutput,
    assertValidOutput,
    getInitInputs,
    lintConnections,
    makeQuantum,
    runTestQuantum,
)


class VisitConnections(PipelineTaskConnections, dimensions={"instrument", "visit"}):
    initIn = connectionTypes.InitInput(
        name="VisitInitIn",
        storageClass="StructuredData",
        multiple=False,
    )
    a = connectionTypes.Input(
        name="VisitA",
        storageClass="StructuredData",
        multiple=False,
        dimensions={"instrument", "visit"},
    )
    b = connectionTypes.Input(
        name="VisitB",
        storageClass="StructuredData",
        multiple=False,
        dimensions={"instrument", "visit"},
    )
    initOut = connectionTypes.InitOutput(
        name="VisitInitOut",
        storageClass="StructuredData",
        multiple=True,
    )
    outA = connectionTypes.Output(
        name="VisitOutA",
        storageClass="StructuredData",
        multiple=False,
        dimensions={"instrument", "visit"},
    )
    outB = connectionTypes.Output(
        name="VisitOutB",
        storageClass="StructuredData",
        multiple=False,
        dimensions={"instrument", "visit"},
    )

    def __init__(self, *, config=None):
        super().__init__(config=config)

        if not config.doUseInitIn:
            self.initInputs.remove("initIn")


class PatchConnections(PipelineTaskConnections, dimensions={"skymap", "tract"}):
    a = connectionTypes.Input(
        name="PatchA",
        storageClass="StructuredData",
        multiple=True,
        dimensions={"skymap", "tract", "patch"},
    )
    b = connectionTypes.PrerequisiteInput(
        name="PatchB",
        storageClass="StructuredData",
        multiple=False,
        dimensions={"skymap", "tract"},
    )
    initOutA = connectionTypes.InitOutput(
        name="PatchInitOutA",
        storageClass="StructuredData",
        multiple=False,
    )
    initOutB = connectionTypes.InitOutput(
        name="PatchInitOutB",
        storageClass="StructuredData",
        multiple=False,
    )
    out = connectionTypes.Output(
        name="PatchOut",
        storageClass="StructuredData",
        multiple=True,
        dimensions={"skymap", "tract", "patch"},
    )

    def __init__(self, *, config=None):
        super().__init__(config=config)

        if not config.doUseB:
            self.prerequisiteInputs.remove("b")


class SkyPixConnections(PipelineTaskConnections, dimensions={"skypix"}):
    a = connectionTypes.Input(
        name="PixA",
        storageClass="StructuredData",
        dimensions={"skypix"},
    )
    out = connectionTypes.Output(
        name="PixOut",
        storageClass="StructuredData",
        dimensions={"skypix"},
    )


class VisitConfig(PipelineTaskConfig, pipelineConnections=VisitConnections):
    doUseInitIn = lsst.pex.config.Field(default=False, dtype=bool, doc="")


class PatchConfig(PipelineTaskConfig, pipelineConnections=PatchConnections):
    doUseB = lsst.pex.config.Field(default=True, dtype=bool, doc="")


class SkyPixConfig(PipelineTaskConfig, pipelineConnections=SkyPixConnections):
    pass


class VisitTask(PipelineTask):
    ConfigClass = VisitConfig
    _DefaultName = "visit"

    def __init__(self, initInputs=None, **kwargs):
        super().__init__(initInputs=initInputs, **kwargs)
        self.initOut = [
            butlerTests.MetricsExample(data=[1, 2]),
            butlerTests.MetricsExample(data=[3, 4]),
        ]

    def run(self, a, b):
        outA = butlerTests.MetricsExample(data=(a.data + b.data))
        outB = butlerTests.MetricsExample(data=(a.data * max(b.data)))
        return Struct(outA=outA, outB=outB)


class PatchTask(PipelineTask):
    ConfigClass = PatchConfig
    _DefaultName = "patch"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.initOutA = butlerTests.MetricsExample(data=[1, 2, 4])
        self.initOutB = butlerTests.MetricsExample(data=[1, 2, 3])

    def run(self, a, b=None):
        if self.config.doUseB:
            out = [butlerTests.MetricsExample(data=(oneA.data + b.data)) for oneA in a]
        else:
            out = a
        return Struct(out=out)


class SkyPixTask(PipelineTask):
    ConfigClass = SkyPixConfig
    _DefaultName = "skypix"

    def run(self, a):
        return Struct(out=a)


class PipelineTaskTestSuite(lsst.utils.tests.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # Repository should be re-created for each test case, but
        # this has a prohibitive run-time cost at present
        cls.root = tempfile.mkdtemp()

        cls.repo = butlerTests.makeTestRepo(cls.root)
        butlerTests.addDataIdValue(cls.repo, "instrument", "notACam")
        butlerTests.addDataIdValue(cls.repo, "visit", 101)
        butlerTests.addDataIdValue(cls.repo, "visit", 102)
        butlerTests.addDataIdValue(cls.repo, "skymap", "sky")
        butlerTests.addDataIdValue(cls.repo, "tract", 42)
        butlerTests.addDataIdValue(cls.repo, "patch", 0)
        butlerTests.addDataIdValue(cls.repo, "patch", 1)
        butlerTests.registerMetricsExample(cls.repo)

        for typeName in {"VisitA", "VisitB", "VisitOutA", "VisitOutB"}:
            butlerTests.addDatasetType(cls.repo, typeName, {"instrument", "visit"}, "StructuredData")
        for typeName in {"PatchA", "PatchOut"}:
            butlerTests.addDatasetType(cls.repo, typeName, {"skymap", "tract", "patch"}, "StructuredData")
        butlerTests.addDatasetType(cls.repo, "PatchB", {"skymap", "tract"}, "StructuredData")
        for typeName in {"PixA", "PixOut"}:
            butlerTests.addDatasetType(cls.repo, typeName, {"htm7"}, "StructuredData")
        butlerTests.addDatasetType(cls.repo, "VisitInitIn", set(), "StructuredData")

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.root, ignore_errors=True)
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.butler = butlerTests.makeTestCollection(self.repo)

    def _makeVisitTestData(self, dataId):
        """Create dummy datasets suitable for VisitTask.

        This method updates ``self.butler`` directly.

        Parameters
        ----------
        dataId : any data ID type
            The (shared) ID for the datasets to create.

        Returns
        -------
        datasets : `dict` [`str`, `list`]
            A dictionary keyed by dataset type. Its values are the list of
            integers used to create each dataset. The datasets stored in the
            butler are `lsst.daf.butler.tests.MetricsExample` objects with
            these lists as their ``data`` argument, but the lists are easier
            to manipulate in test code.
        """
        inInit = [4, 2]
        inA = [1, 2, 3]
        inB = [4, 0, 1]
        self.butler.put(butlerTests.MetricsExample(data=inA), "VisitA", dataId)
        self.butler.put(butlerTests.MetricsExample(data=inB), "VisitB", dataId)
        self.butler.put(butlerTests.MetricsExample(data=inInit), "VisitInitIn", set())
        return {
            "VisitA": inA,
            "VisitB": inB,
            "VisitInitIn": inInit,
        }

    def _makePatchTestData(self, dataId):
        """Create dummy datasets suitable for PatchTask.

        This method updates ``self.butler`` directly.

        Parameters
        ----------
        dataId : any data ID type
            The (shared) ID for the datasets to create. Any patch ID is
            overridden to create multiple datasets.

        Returns
        -------
        datasets : `dict` [`str`, `list` [`tuple` [data ID, `list`]]]
            A dictionary keyed by dataset type. Its values are the data ID
            of each dataset and the list of integers used to create each. The
            datasets stored in the butler are
            `lsst.daf.butler.tests.MetricsExample` objects with these lists as
            their ``data`` argument, but the lists are easier to manipulate
            in test code.
        """
        inA = [1, 2, 3]
        inB = [4, 0, 1]
        datasets = {"PatchA": [], "PatchB": []}
        for patch in {0, 1}:
            self.butler.put(butlerTests.MetricsExample(data=(inA + [patch])), "PatchA", dataId, patch=patch)
            datasets["PatchA"].append((dict(dataId, patch=patch), inA + [patch]))
        self.butler.put(butlerTests.MetricsExample(data=inB), "PatchB", dataId)
        datasets["PatchB"].append((dataId, inB))
        return datasets

    def testMakeQuantumNoSuchDatatype(self):
        config = VisitConfig()
        config.connections.a = "Visit"
        task = VisitTask(config=config)

        dataId = {"instrument": "notACam", "visit": 102}
        self._makeVisitTestData(dataId)

        with self.assertRaises(ValueError):
            makeQuantum(task, self.butler, dataId, {key: dataId for key in {"a", "b", "outA", "outB"}})

    def testMakeQuantumInvalidDimension(self):
        config = VisitConfig()
        config.connections.a = "PatchA"
        task = VisitTask(config=config)
        dataIdV = {"instrument": "notACam", "visit": 102}
        dataIdVExtra = {"instrument": "notACam", "visit": 102, "detector": 42}
        dataIdP = {"skymap": "sky", "tract": 42, "patch": 0}

        inA = [1, 2, 3]
        inB = [4, 0, 1]
        self.butler.put(butlerTests.MetricsExample(data=inA), "VisitA", dataIdV)
        self.butler.put(butlerTests.MetricsExample(data=inA), "PatchA", dataIdP)
        self.butler.put(butlerTests.MetricsExample(data=inB), "VisitB", dataIdV)

        # dataIdV is correct everywhere, dataIdP should error
        with self.assertRaises(ValueError):
            makeQuantum(
                task,
                self.butler,
                dataIdV,
                {
                    "a": dataIdP,
                    "b": dataIdV,
                    "outA": dataIdV,
                    "outB": dataIdV,
                },
            )
        with self.assertRaises(ValueError):
            makeQuantum(
                task,
                self.butler,
                dataIdP,
                {
                    "a": dataIdV,
                    "b": dataIdV,
                    "outA": dataIdV,
                    "outB": dataIdV,
                },
            )
        # should not accept small changes, either
        with self.assertRaises(ValueError):
            makeQuantum(
                task,
                self.butler,
                dataIdV,
                {
                    "a": dataIdV,
                    "b": dataIdV,
                    "outA": dataIdVExtra,
                    "outB": dataIdV,
                },
            )
        with self.assertRaises(ValueError):
            makeQuantum(
                task,
                self.butler,
                dataIdVExtra,
                {
                    "a": dataIdV,
                    "b": dataIdV,
                    "outA": dataIdV,
                    "outB": dataIdV,
                },
            )

    def testMakeQuantumMissingMultiple(self):
        task = PatchTask()

        dataId = {"skymap": "sky", "tract": 42}
        self._makePatchTestData(dataId)

        with self.assertRaises(ValueError):
            makeQuantum(
                task,
                self.butler,
                dataId,
                {
                    "a": dict(dataId, patch=0),
                    "b": dataId,
                    "out": [dict(dataId, patch=patch) for patch in {0, 1}],
                },
            )

    def testMakeQuantumExtraMultiple(self):
        task = PatchTask()

        dataId = {"skymap": "sky", "tract": 42}
        self._makePatchTestData(dataId)

        with self.assertRaises(ValueError):
            makeQuantum(
                task,
                self.butler,
                dataId,
                {
                    "a": [dict(dataId, patch=patch) for patch in {0, 1}],
                    "b": [dataId],
                    "out": [dict(dataId, patch=patch) for patch in {0, 1}],
                },
            )

    def testMakeQuantumMissingDataId(self):
        task = VisitTask()

        dataId = {"instrument": "notACam", "visit": 102}
        self._makeVisitTestData(dataId)

        with self.assertRaises(ValueError):
            makeQuantum(task, self.butler, dataId, {key: dataId for key in {"a", "outA", "outB"}})
        with self.assertRaises(ValueError):
            makeQuantum(task, self.butler, dataId, {key: dataId for key in {"a", "b", "outB"}})

    def testMakeQuantumCorruptedDataId(self):
        task = VisitTask()

        dataId = {"instrument": "notACam", "visit": 102}
        self._makeVisitTestData(dataId)

        with self.assertRaises(ValueError):
            # fourth argument should be a mapping keyed by component name
            makeQuantum(task, self.butler, dataId, dataId)

    def testRunTestQuantumVisitWithRun(self):
        task = VisitTask()

        dataId = {"instrument": "notACam", "visit": 102}
        data = self._makeVisitTestData(dataId)

        quantum = makeQuantum(task, self.butler, dataId, {key: dataId for key in {"a", "b", "outA", "outB"}})
        runTestQuantum(task, self.butler, quantum, mockRun=False)

        # Can we use runTestQuantum to verify that task.run got called with
        # correct inputs/outputs?
        self.assertTrue(self.butler.datasetExists("VisitOutA", dataId))
        self.assertEqual(
            self.butler.get("VisitOutA", dataId),
            butlerTests.MetricsExample(data=(data["VisitA"] + data["VisitB"])),
        )
        self.assertTrue(self.butler.datasetExists("VisitOutB", dataId))
        self.assertEqual(
            self.butler.get("VisitOutB", dataId),
            butlerTests.MetricsExample(data=(data["VisitA"] * max(data["VisitB"]))),
        )

    def testRunTestQuantumPatchWithRun(self):
        task = PatchTask()

        dataId = {"skymap": "sky", "tract": 42}
        data = self._makePatchTestData(dataId)

        quantum = makeQuantum(
            task,
            self.butler,
            dataId,
            {
                "a": [dataset[0] for dataset in data["PatchA"]],
                "b": dataId,
                "out": [dataset[0] for dataset in data["PatchA"]],
            },
        )
        runTestQuantum(task, self.butler, quantum, mockRun=False)

        # Can we use runTestQuantum to verify that task.run got called with
        # correct inputs/outputs?
        inB = data["PatchB"][0][1]
        for dataset in data["PatchA"]:
            patchId = dataset[0]
            self.assertTrue(self.butler.datasetExists("PatchOut", patchId))
            self.assertEqual(
                self.butler.get("PatchOut", patchId), butlerTests.MetricsExample(data=(dataset[1] + inB))
            )

    def testRunTestQuantumVisitMockRun(self):
        task = VisitTask()

        dataId = {"instrument": "notACam", "visit": 102}
        data = self._makeVisitTestData(dataId)

        quantum = makeQuantum(task, self.butler, dataId, {key: dataId for key in {"a", "b", "outA", "outB"}})
        run = runTestQuantum(task, self.butler, quantum, mockRun=True)

        # Can we use the mock to verify that task.run got called with the
        # correct inputs?
        run.assert_called_once_with(
            a=butlerTests.MetricsExample(data=data["VisitA"]),
            b=butlerTests.MetricsExample(data=data["VisitB"]),
        )

    def testRunTestQuantumPatchMockRun(self):
        task = PatchTask()

        dataId = {"skymap": "sky", "tract": 42}
        data = self._makePatchTestData(dataId)

        quantum = makeQuantum(
            task,
            self.butler,
            dataId,
            {
                # Use lists, not sets, to ensure order agrees with test
                # assertion.
                "a": [dataset[0] for dataset in data["PatchA"]],
                "b": dataId,
                "out": [dataset[0] for dataset in data["PatchA"]],
            },
        )
        run = runTestQuantum(task, self.butler, quantum, mockRun=True)

        # Can we use the mock to verify that task.run got called with the
        # correct inputs?
        run.assert_called_once_with(
            a=[butlerTests.MetricsExample(data=dataset[1]) for dataset in data["PatchA"]],
            b=butlerTests.MetricsExample(data=data["PatchB"][0][1]),
        )

    def testRunTestQuantumPatchOptionalInput(self):
        config = PatchConfig()
        config.doUseB = False
        task = PatchTask(config=config)

        dataId = {"skymap": "sky", "tract": 42}
        data = self._makePatchTestData(dataId)

        quantum = makeQuantum(
            task,
            self.butler,
            dataId,
            {
                # Use lists, not sets, to ensure order agrees with test
                # assertion.
                "a": [dataset[0] for dataset in data["PatchA"]],
                "out": [dataset[0] for dataset in data["PatchA"]],
            },
        )
        run = runTestQuantum(task, self.butler, quantum, mockRun=True)

        # Can we use the mock to verify that task.run got called with the
        # correct inputs?
        run.assert_called_once_with(
            a=[butlerTests.MetricsExample(data=dataset[1]) for dataset in data["PatchA"]]
        )

    def testAssertValidOutputPass(self):
        task = VisitTask()

        inA = butlerTests.MetricsExample(data=[1, 2, 3])
        inB = butlerTests.MetricsExample(data=[4, 0, 1])
        result = task.run(inA, inB)

        # should not throw
        assertValidOutput(task, result)

    def testAssertValidOutputMissing(self):
        task = VisitTask()

        def run(a, b):
            return Struct(outA=a)

        task.run = run

        inA = butlerTests.MetricsExample(data=[1, 2, 3])
        inB = butlerTests.MetricsExample(data=[4, 0, 1])
        result = task.run(inA, inB)

        with self.assertRaises(AssertionError):
            assertValidOutput(task, result)

    def testAssertValidOutputSingle(self):
        task = PatchTask()

        def run(a, b):
            return Struct(out=b)

        task.run = run

        inA = butlerTests.MetricsExample(data=[1, 2, 3])
        inB = butlerTests.MetricsExample(data=[4, 0, 1])
        result = task.run([inA], inB)

        with self.assertRaises(AssertionError):
            assertValidOutput(task, result)

    def testAssertValidOutputMultiple(self):
        task = VisitTask()

        def run(a, b):
            return Struct(outA=[a], outB=b)

        task.run = run

        inA = butlerTests.MetricsExample(data=[1, 2, 3])
        inB = butlerTests.MetricsExample(data=[4, 0, 1])
        result = task.run(inA, inB)

        with self.assertRaises(AssertionError):
            assertValidOutput(task, result)

    def testAssertValidInitOutputPass(self):
        task = VisitTask()
        # should not throw
        assertValidInitOutput(task)

        task = PatchTask()
        # should not throw
        assertValidInitOutput(task)

    def testAssertValidInitOutputMissing(self):
        class BadVisitTask(VisitTask):
            def __init__(self, **kwargs):
                PipelineTask.__init__(self, **kwargs)  # Bypass VisitTask constructor
                pass  # do not set fields

        task = BadVisitTask()

        with self.assertRaises(AssertionError):
            assertValidInitOutput(task)

    def testAssertValidInitOutputSingle(self):
        class BadVisitTask(VisitTask):
            def __init__(self, **kwargs):
                PipelineTask.__init__(self, **kwargs)  # Bypass VisitTask constructor
                self.initOut = butlerTests.MetricsExample(data=[1, 2])

        task = BadVisitTask()

        with self.assertRaises(AssertionError):
            assertValidInitOutput(task)

    def testAssertValidInitOutputMultiple(self):
        class BadPatchTask(PatchTask):
            def __init__(self, **kwargs):
                PipelineTask.__init__(self, **kwargs)  # Bypass PatchTask constructor
                self.initOutA = [butlerTests.MetricsExample(data=[1, 2, 4])]
                self.initOutB = butlerTests.MetricsExample(data=[1, 2, 3])

        task = BadPatchTask()

        with self.assertRaises(AssertionError):
            assertValidInitOutput(task)

    def testGetInitInputs(self):
        dataId = {"instrument": "notACam", "visit": 102}
        data = self._makeVisitTestData(dataId)

        self.assertEqual(getInitInputs(self.butler, VisitConfig()), {})

        config = VisitConfig()
        config.doUseInitIn = True
        self.assertEqual(
            getInitInputs(self.butler, config),
            {"initIn": butlerTests.MetricsExample(data=data["VisitInitIn"])},
        )

    def testSkypixHandling(self):
        task = SkyPixTask()

        dataId = {"htm7": 157227}  # connection declares skypix, but Butler uses htm7
        data = butlerTests.MetricsExample(data=[1, 2, 3])
        self.butler.put(data, "PixA", dataId)

        quantum = makeQuantum(task, self.butler, dataId, {key: dataId for key in {"a", "out"}})
        run = runTestQuantum(task, self.butler, quantum, mockRun=True)

        # PixA dataset should have been retrieved by runTestQuantum
        run.assert_called_once_with(a=data)

    def testLintConnectionsOk(self):
        lintConnections(VisitConnections)
        lintConnections(PatchConnections)
        lintConnections(SkyPixConnections)

    def testLintConnectionsMissingMultiple(self):
        class BadConnections(PipelineTaskConnections, dimensions={"tract", "patch", "skymap"}):
            coadds = connectionTypes.Input(
                name="coadd_calexp",
                storageClass="ExposureF",
                # Some authors use list rather than set; check that linter
                # can handle it.
                dimensions=["tract", "patch", "band", "skymap"],
            )

        with self.assertRaises(AssertionError):
            lintConnections(BadConnections)
        lintConnections(BadConnections, checkMissingMultiple=False)

    def testLintConnectionsExtraMultiple(self):
        class BadConnections(
            PipelineTaskConnections,
            # Some authors use list rather than set.
            dimensions=["tract", "patch", "band", "skymap"],
        ):
            coadds = connectionTypes.Input(
                name="coadd_calexp",
                storageClass="ExposureF",
                multiple=True,
                dimensions={"tract", "patch", "band", "skymap"},
            )

        with self.assertRaises(AssertionError):
            lintConnections(BadConnections)
        lintConnections(BadConnections, checkUnnecessaryMultiple=False)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    pass


def setup_module(module):
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any

import click
from lsst.daf.butler.cli.opt import repo_argument
from lsst.daf.butler.cli.utils import ButlerCommand

from ... import script
from ..opt import instrument_argument


@click.command(short_help="Add an instrument definition to the repository", cls=ButlerCommand)
@repo_argument(required=True)
@instrument_argument(required=True, nargs=-1, help="The fully-qualified name of an Instrument subclass.")
@click.option("--update", is_flag=True)
def register_instrument(*args: Any, **kwargs: Any) -> None:
    """Add an instrument to the data repository."""
    script.register_instrument(*args, **kwargs)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import uuid

__all__ = ("QuantumNode", "NodeId", "BuildId")

from dataclasses import dataclass
from typing import Any, Dict, NewType, Optional, Tuple

from lsst.daf.butler import (
    DatasetRef,
    DimensionRecord,
    DimensionRecordsAccumulator,
    DimensionUniverse,
    Quantum,
    SerializedQuantum,
)
from pydantic import BaseModel

from ..pipeline import TaskDef

BuildId = NewType("BuildId", str)


def _hashDsRef(ref: DatasetRef) -> int:
    return hash((ref.datasetType, ref.dataId))


@dataclass(frozen=True, eq=True)
class NodeId:
    """Deprecated, this class is used with QuantumGraph save formats of
    1 and 2 when unpicking objects and must be retained until those formats
    are considered unloadable.

    This represents an unique identifier of a node within an individual
    construction of a `QuantumGraph`. This identifier will stay constant
    through a pickle, and any `QuantumGraph` methods that return a new
    `QuantumGraph`.

    A `NodeId` will not be the same if a new graph is built containing the same
    information in a `QuantumNode`, or even built from exactly the same inputs.

    `NodeId`s do not play any role in deciding the equality or identity (hash)
    of a `QuantumNode`, and are mainly useful in debugging or working with
    various subsets of the same graph.

    This interface is a convenance only, and no guarantees on long term
    stability are made. New implementations might change the `NodeId`, or
    provide more or less guarantees.
    """

    number: int
    """The unique position of the node within the graph assigned at graph
    creation.
    """
    buildId: BuildId
    """Unique identifier created at the time the originating graph was created
    """


@dataclass(frozen=True)
class QuantumNode:
    """This class represents a node in the quantum graph.

    The quantum attribute represents the data that is to be processed at this
    node.
    """

    quantum: Quantum
    """The unit of data that is to be processed by this graph node"""
    taskDef: TaskDef
    """Definition of the task that will process the `Quantum` associated with
    this node.
    """
    nodeId: uuid.UUID
    """The unique position of the node within the graph assigned at graph
    creation.
    """

    def __post_init__(self) -> None:
        # use setattr here to preserve the frozenness of the QuantumNode
        self._precomputedHash: int
        object.__setattr__(self, "_precomputedHash", hash((self.taskDef.label, self.quantum)))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, QuantumNode):
            return False
        if self.quantum != other.quantum:
            return False
        return self.taskDef == other.taskDef

    def __hash__(self) -> int:
        """For graphs it is useful to have a more robust hash than provided
        by the default quantum id based hashing
        """
        return self._precomputedHash

    def __repr__(self) -> str:
        """Make more human readable string representation."""
        return (
            f"{self.__class__.__name__}(quantum={self.quantum}, "
            f"taskDef={self.taskDef}, nodeId={self.nodeId})"
        )

    def to_simple(self, accumulator: Optional[DimensionRecordsAccumulator] = None) -> SerializedQuantumNode:
        return SerializedQuantumNode(
            quantum=self.quantum.to_simple(accumulator=accumulator),
            taskLabel=self.taskDef.label,
            nodeId=self.nodeId,
        )

    @classmethod
    def from_simple(
        cls,
        simple: SerializedQuantumNode,
        taskDefMap: Dict[str, TaskDef],
        universe: DimensionUniverse,
        recontitutedDimensions: Optional[Dict[int, Tuple[str, DimensionRecord]]] = None,
    ) -> QuantumNode:
        return QuantumNode(
            quantum=Quantum.from_simple(
                simple.quantum, universe, reconstitutedDimensions=recontitutedDimensions
            ),
            taskDef=taskDefMap[simple.taskLabel],
            nodeId=simple.nodeId,
        )


class SerializedQuantumNode(BaseModel):
    quantum: SerializedQuantum
    taskLabel: str
    nodeId: uuid.UUID

    @classmethod
    def direct(cls, *, quantum: Dict[str, Any], taskLabel: str, nodeId: str) -> SerializedQuantumNode:
        node = SerializedQuantumNode.__new__(cls)
        setter = object.__setattr__
        setter(node, "quantum", SerializedQuantum.direct(**quantum))
        setter(node, "taskLabel", taskLabel)
        setter(node, "nodeId", uuid.UUID(nodeId))
        setter(node, "__fields_set__", {"quantum", "taskLabel", "nodeId"})
        return node



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for Pipeline.
"""

import pickle
import textwrap
import unittest

import lsst.utils.tests
from lsst.pipe.base import Pipeline, PipelineDatasetTypes, TaskDef
from lsst.pipe.base.tests.simpleQGraph import AddTask, makeSimplePipeline


class PipelineTestCase(unittest.TestCase):
    """A test case for TaskDef and Pipeline."""

    def testTaskDef(self):
        """Tests for TaskDef structure"""
        task1 = TaskDef(taskClass=AddTask, config=AddTask.ConfigClass())
        self.assertIn("Add", task1.taskName)
        self.assertIsInstance(task1.config, AddTask.ConfigClass)
        self.assertIsNotNone(task1.taskClass)
        self.assertEqual(task1.label, "add_task")
        task1a = pickle.loads(pickle.dumps(task1))
        self.assertEqual(task1, task1a)

    def testEmpty(self):
        """Creating empty pipeline"""
        pipeline = Pipeline("test")
        self.assertEqual(len(pipeline), 0)

    def testInitial(self):
        """Testing constructor with initial data"""
        pipeline = makeSimplePipeline(2)
        self.assertEqual(len(pipeline), 2)
        expandedPipeline = list(pipeline.toExpandedPipeline())
        self.assertEqual(expandedPipeline[0].taskName, "lsst.pipe.base.tests.simpleQGraph.AddTask")
        self.assertEqual(expandedPipeline[1].taskName, "lsst.pipe.base.tests.simpleQGraph.AddTask")
        self.assertEqual(expandedPipeline[0].taskClass, AddTask)
        self.assertEqual(expandedPipeline[1].taskClass, AddTask)
        self.assertEqual(expandedPipeline[0].label, "task0")
        self.assertEqual(expandedPipeline[1].label, "task1")

    def testParameters(self):
        """Test that parameters can be set and used to format"""
        pipeline_str = textwrap.dedent(
            """
            description: Test Pipeline
            parameters:
               testValue: 5
            tasks:
              add:
                class: test_pipeline.AddTask
                config:
                  addend: parameters.testValue
        """
        )
        # verify that parameters are used in expanding a pipeline
        pipeline = Pipeline.fromString(pipeline_str)
        expandedPipeline = list(pipeline.toExpandedPipeline())
        self.assertEqual(expandedPipeline[0].config.addend, 5)

        # verify that a parameter can be overridden on the "command line"
        pipeline.addConfigOverride("parameters", "testValue", 14)
        expandedPipeline = list(pipeline.toExpandedPipeline())
        self.assertEqual(expandedPipeline[0].config.addend, 14)

        # verify that a non existing parameter cant be overridden
        with self.assertRaises(ValueError):
            pipeline.addConfigOverride("parameters", "missingValue", 17)

        # verify that parameters does not support files or python overrides
        with self.assertRaises(ValueError):
            pipeline.addConfigFile("parameters", "fakeFile")
        with self.assertRaises(ValueError):
            pipeline.addConfigPython("parameters", "fakePythonString")

    def testSerialization(self):
        pipeline = makeSimplePipeline(2)
        dump = str(pipeline)
        load = Pipeline.fromString(dump)
        self.assertEqual(pipeline, load)

    def test_initOutputNames(self):
        """Test for PipelineDatasetTypes.initOutputNames method."""
        pipeline = makeSimplePipeline(3)
        dsType = set(PipelineDatasetTypes.initOutputNames(pipeline))
        expected = {
            "packages",
            "add_init_output1",
            "add_init_output2",
            "add_init_output3",
            "task0_config",
            "task1_config",
            "task2_config",
        }
        self.assertEqual(dsType, expected)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    pass


def setup_module(module):
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



from . import connectionTypes, pipelineIR
from ._instrument import *
from ._status import *
from ._task_metadata import *

try:
    from .argumentParser import *
except ImportError:
    # Gen2 imports are optional.
    pass
from .butlerQuantumContext import *

try:
    from .cmdLineTask import *
except ImportError:
    # Gen2 imports are optional.
    pass
from .config import *
from .connections import *
from .executionButlerBuilder import *
from .graph import *
from .graphBuilder import *
from .pipeline import *
from .pipelineTask import *
from .struct import *
from .task import *
from .task_logging import getTaskLogger
from .taskFactory import *
from .timer import *
from .version import *



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for Task logging.
"""

import logging
import unittest

from lsst.pipe.base import getTaskLogger


class TestLogging(unittest.TestCase):
    def testLogLevels(self):
        """Check that the new log levels look reasonable."""

        root = getTaskLogger()

        self.assertEqual(root.DEBUG, logging.DEBUG)
        self.assertGreater(root.VERBOSE, logging.DEBUG)
        self.assertLess(root.VERBOSE, logging.INFO)
        self.assertLess(root.TRACE, logging.DEBUG)

    def testLogCommands(self):
        """Check that all the log commands work."""

        root = getTaskLogger()

        with self.assertLogs(level=root.TRACE) as cm:
            root.trace("Trace")
            root.debug("Debug")
            root.verbose("Verbose")
            root.info("Info")
            root.warning("Warning")
            root.fatal("Fatal")
            root.critical("Critical")
            root.error("Error")

        self.assertEqual(len(cm.records), 8)

        # Check that each record has an explicit level name rather than
        # "Level N" and comes from this file (and not the task_logging.py).
        for record in cm.records:
            self.assertRegex(record.levelname, "^[A-Z]+$")
            self.assertEqual(record.filename, "test_logging.py")


if __name__ == "__main__":
    unittest.main()
